# RCAIDE/Library/Plots/Noise/plot_3D_noise_contour.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots  import *

# python imports 
import numpy as np  
import plotly.graph_objects as go

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------    
def plot_3D_noise_contour(noise_data,
                       noise_level              = None ,
                       min_noise_level          = 35,  
                       max_noise_level          = 90, 
                       noise_scale_label        = None,
                       save_figure              = False,
                       show_figure              = True,
                       save_filename            = "Noise_Contour",
                       use_lat_long_coordinates = True, 
                       show_trajectory          = False,
                       show_microphones         = False,
                       colormap                 = 'jet',
                       file_type                = ".png",
                       background_color         = 'rgb(17,54,71)',
                       grid_color               = 'gray',
                       width                    = 1400, 
                       height                   = 800,
                       *args, **kwargs): 
    """This plots a 3D noise contour of a noise level using plotly

    Assumptions:
    None

    Source:
    None

    Inputs: 
       noise_data        - noise data structure 
       noise_level       - noise level (dBA, DNL, SENEL etc)
       min_noise_level   - minimal noise level 
       max_noise_level   - maximum noise level 
       noise_scale_label - noise level label 
       save_figure       - save figure 
       show_figure       - show figure 
       save_filename     - save file flag
       show_trajectory   - plot aircraft trajectory flag
       show_microphones  - show microhpone flag 

    Outputs:
       Plots

    Properties Used:
    N/A
    """   
    Aircraft_pos    = noise_data.aircraft_position      
    X               = noise_data.microphone_locations[:,:,0]/Units.nmi  
    Y               = noise_data.microphone_locations[:,:,1]/Units.nmi  
    Z               = noise_data.microphone_locations[:,:,2]/Units.feet  
    plot_data       = []   
  
    # ---------------------------------------------------------------------------
    # TRHEE DIMENSIONAL NOISE CONTOUR
    # --------------------------------------------------------------------------- 
    # TERRAIN CONTOUR 
    ground_contour   = contour_surface_slice(Y,X,Z,noise_level,color_scale=colormap)
    plot_data.append(ground_contour)

    # GROUND MICROPHONES
    if show_microphones:
        microphones = go.Scatter3d(x        = Y.flatten(),
                                   y        = X.flatten(),
                                   z        = Z.flatten(),
                                   mode     = 'markers',
                                   marker   = dict(size=6,color='white',opacity=0.8),
                                   line     = dict(color='white',width=2))
        plot_data.append(microphones)

    # AIRCRAFT TRAJECTORY
    if show_trajectory:
        aircraft_trajectory = go.Scatter3d(x   = Aircraft_pos[:,1]/Units.nmi,
                                           y   = Aircraft_pos[:,0]/Units.nmi,
                                           z   = Aircraft_pos[:,2]/Units.feet,
                                           mode= 'markers',
                                           marker=dict(size=6,
                                                       color='black',
                                                       opacity=0.8),
                                    line=dict(color='black',width=2))
        plot_data.append(aircraft_trajectory)

    # Define Colorbar Bounds
    min_alt     = np.minimum(np.min(Z),0) 
    max_alt     = np.maximum(np.max(Z), np.max(Aircraft_pos[:,2]/Units.feet)) 
  
    fig_3d = go.Figure(data=plot_data) 

    if show_microphones or show_trajectory:
        pass
    else: 
        fig_3d.update_traces(colorbar_orientation     = 'v',
                             colorbar_thickness       = 50,
                             colorbar_nticks          = 10,
                             colorbar_title_text      = noise_scale_label,
                             colorbar_tickfont_size   = 16,
                             colorbar_title_side      = "right",
                             colorbar_ypad            = 50,
                             colorbar_len             = 0.75)
        
                         
    fig_3d.update_layout(
             title_text                             = save_filename, 
             title_x                                = 0.5,
             width                                  = width,
             height                                 = height,
             font_size                              = 12,
             scene_aspectmode                       = "manual",
             scene_aspectratio                      = dict(x=1, y=1, z=0.5),      
             scene_zaxis_range                      = [min_alt,max_alt*3],
             scene                                  = dict(xaxis_title='Latitude [nmi]',
                                                           yaxis_title='Longitude [nmi]',
                                                           zaxis_title='Elevation [ft]',
                                                           xaxis = dict(
                                                                backgroundcolor=background_color,
                                                                gridcolor="white",
                                                                showbackground=True,
                                                                zerolinecolor=grid_color,),
                                                           yaxis = dict(
                                                               backgroundcolor=background_color,
                                                               gridcolor=grid_color,
                                                               showbackground=True,
                                                               zerolinecolor="white"),
                                                           zaxis = dict(
                                                               backgroundcolor=background_color,
                                                               gridcolor=grid_color,
                                                               showbackground=True,
                                                               zerolinecolor="white",),),
             scene_camera=dict(up    = dict(x=0, y=0, z=1),
                               center= dict(x=-0.05, y=0, z=-0.20),
                               eye   = dict(x=-1.0, y=-1.0, z=.4))   
    ) 
    if show_figure:
        fig_3d.show() 
    if save_figure:
        fig_3d.write_image(save_filename, file_type)

    return fig_3d       

def colorax(vmin, vmax):
    return dict(cmin=vmin, cmax=vmax)
 


# RCAIDE/Library/Methods/Propulsors/Converters/Rotor/compute_ducted_fan_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports 
from RCAIDE.Framework.Core                              import Data , Units, orientation_product, orientation_transpose   

# package imports
import  numpy as  np 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Generalized Rotor Class
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_ducted_fan_performance(propulsor,state,center_of_gravity= [[0.0, 0.0,0.0]]):
    """Analyzes a general ducted_fan given geometry and operating conditions.

    Assumptions:
    N.A.

    Source:
    N.A.
    
    Inputs:
        propulsor          (dict): propulsor data structure 
        state              (dict): flight conditions data structure  
        center_of_gravity  (list): center of gravity  

    Outputs:
        None
    """

    # Unpack ducted_fan blade parameters and operating conditions 
    conditions            = state.conditions
    ducted_fan            = propulsor.ducted_fan
    propulsor_conditions  = conditions.energy[propulsor.tag]
    commanded_TV          = propulsor_conditions.commanded_thrust_vector_angle
    ducted_fan_conditions = propulsor_conditions[ducted_fan.tag]
                  
    altitude  = conditions.freestream.altitude / 1000
    a         = conditions.freestream.speed_of_sound
    
    omega = ducted_fan_conditions.omega 
    n     = omega/(2.*np.pi)   # Rotations per second
    D     = ducted_fan.tip_radius * 2
    A     = 0.25 * np.pi * (D ** 2)
    
    # Unpack freestream conditions
    rho     = conditions.freestream.density[:,0,None] 
    Vv      = conditions.frames.inertial.velocity_vector 

    # Number of radial stations and segment control point
    B        = ducted_fan.number_of_rotor_blades
    Nr       = ducted_fan.number_of_radial_stations
    ctrl_pts = len(Vv)
     
    # Velocity in the rotor frame
    T_body2inertial         = conditions.frames.body.transform_to_inertial
    T_inertial2body         = orientation_transpose(T_body2inertial)
    V_body                  = orientation_product(T_inertial2body,Vv)
    body2thrust,orientation = ducted_fan.body_to_prop_vel(commanded_TV) 
    T_body2thrust           = orientation_transpose(np.ones_like(T_body2inertial[:])*body2thrust)
    V_thrust                = orientation_product(T_body2thrust,V_body)

    # Check and correct for hover
    V         = V_thrust[:,0,None]
    V[V==0.0] = 1E-6
     
    tip_mach = (omega * ducted_fan.tip_radius) / a
    mach     =  V/ a
    # create tuple for querying surrogate 
    pts      = (mach,tip_mach,altitude) 
    
    thrust         = ducted_fan.performance_surrogates.thrust(pts)            
    power          = ducted_fan.performance_surrogates.power(pts)                 
    efficiency     = ducted_fan.performance_surrogates.efficiency(pts)            
    torque         = ducted_fan.performance_surrogates.torque(pts)                
    Ct             = ducted_fan.performance_surrogates.thrust_coefficient(pts)    
    Cp             = ducted_fan.performance_surrogates.power_coefficient(pts) 
    Cq             = torque/(rho*(n*n)*(D*D*D*D*D))
    FoM            = thrust*np.sqrt(thrust/(2*rho*A))/power  
    
    # calculate coefficients    
    thrust_prop_frame      = np.zeros((ctrl_pts,3))
    thrust_prop_frame[:,0] = thrust[:,0]
    thrust_vector          = orientation_product(orientation_transpose(T_body2thrust),thrust_prop_frame)
 
    # Compute moment 
    moment_vector           = np.zeros((ctrl_pts,3))
    moment_vector[:,0]      = ducted_fan.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = ducted_fan.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = ducted_fan.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, thrust_vector) 
     
    outputs                                       = Data( 
                torque                            = torque,
                thrust                            = thrust_vector,  
                power                             = power,
                moment                            = moment, 
                rpm                               = omega /Units.rpm ,   
                tip_mach                          = tip_mach, 
                efficiency                        = efficiency,         
                number_radial_stations            = Nr, 
                orientation                       = orientation, 
                speed_of_sound                    = conditions.freestream.speed_of_sound,
                density                           = conditions.freestream.density,
                velocity                          = Vv,     
                omega                             = omega,  
                thrust_per_blade                  = thrust/B,
                thrust_coefficient                = Ct, 
                torque_per_blade                  = torque/B,
                figure_of_merit                   = FoM, 
                torque_coefficient                = Cq,
                power_coefficient                 = Cp,  
        ) 
    
    conditions.energy[propulsor.tag][ducted_fan.tag] = outputs   
    
    return  



## @ingroup Methods-Noise-Multi_Fidelity
# RCAIDE/Methods/Noise/Multi_Fidelity/harmonic_noise_plane.py
# 
# 
# Created:  Jul 2024, Niranjan Nanjappa

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE
from RCAIDE.Framework.Core                                 import orientation_product, orientation_transpose  
from RCAIDE.Library.Methods.Noise.Common                   import convert_to_third_octave_band

# Python Package imports  
import numpy as np
from scipy.special import jv 
import scipy as sp

# ----------------------------------------------------------------------------------------------------------------------
# Compute Harmonic Noise 
# ----------------------------------------------------------------------------------------------------------------------
## @ingroup Methods-Noise-Frequency_Domain_Buildup-Rotor 
def harmonic_noise_plane(harmonics_blade,harmonics_load,conditions,propulsor_conditions,coordinates,rotor,settings,Noise,cpt):
    '''This computes the harmonic noise (i.e. thickness and loading noise) in the frequency domain 
    of a rotor at any angle of attack with load distribution along the blade span and blade chord. This is a 
    level 3 fidelity approach. All sources are computed using the helicoidal surface theory.

    Assumptions:
    1) Acoustic non-compactness of loads along blade chord.
    2) Acoustic non-compactness of loads along blade span.
    3) Acoustic compactness of loads along blade thickness.

    Source:
    1) Hanson, D. B. (1995). Sound from a propeller at angle of attack: a new theoretical viewpoint. 
    Proceedings - Royal Society of London, A, 449(1936).
    
    2) Hanson, D. B. "Noise radiation of propeller loading sources with angular inflow" AIAA 1990-3955.
    13th Aeroacoustics Conference. October 1990.
    
    3) Hanson, Donald B. "Helicoidal surface theory for harmonic noise of rotors in the far field."
    AIAA Journal 18.10 (1980): 1213-1220.

    3) Hubbard, Harvey H., ed. Aeroacoustics of flight vehicles: theory and practice. Vol. 1.
    NASA Office of Management, Scientific and Technical Information Program, 1991.


    Inputs: 
        harmonics_blade               - blade harmonics                                                            [Unitless]
        harmonics_load                - loading harmonics (modes within each blade harmonic mode)                  [Unitless]
        freestream                    - freestream data structure                                                  [m/s]
        angle_of_attack               - aircraft angle of attack                                                   [rad]
        position_vector               - position vector of aircraft                                                [m]
        velocity_vector               - velocity vector of aircraft                                                [m/s] 
        rotors                        - data structure of rotors                                                   [None]
        aeroacoustic_data             - data structure of acoustic data                                            [None]
        settings                      - accoustic settings                                                         [None] 
        res                           - results data structure                                                     [None] 

    Outputs 
        res.                                    *acoustic data is stored and passed in data structures*                                                                            
            SPL_prop_harmonic_bpf_spectrum       - harmonic noise in blade passing frequency spectrum              [dB]
            SPL_prop_harmonic_bpf_spectrum_dBA   - dBA-Weighted harmonic noise in blade passing frequency spectrum [dbA]                  
            SPL_prop_harmonic_1_3_spectrum       - harmonic noise in 1/3 octave spectrum                           [dB]
            SPL_prop_harmonic_1_3_spectrum_dBA   - dBA-Weighted harmonic noise in 1/3 octave spectrum              [dBA] 
            p_pref_harmonic                      - pressure ratio of harmonic noise                                [Unitless]
            p_pref_harmonic_dBA                  - pressure ratio of dBA-weighted harmonic noise                   [Unitless]


    Properties Used:
        N/A
    
    Code Convention - The number in front of a variable name indicates the number of dimensions of the variable.
                      For instance, m_6 is the 6 dimensional harmonic modes variable, m_5 is 5 dimensional harmonic modes variable
    '''     

    aeroacoustic_data       = propulsor_conditions[rotor.tag] 
    angle_of_attack         = np.atleast_2d(conditions.aerodynamics.angles.alpha[cpt]) 
    velocity_vector         = np.atleast_2d(conditions.frames.inertial.velocity_vector[cpt])  
    freestream              = conditions.freestream       
    num_h_b                 = len(harmonics_blade)
    num_h_l                 = len(harmonics_load)
    num_cpt                 = len(angle_of_attack) 
    num_mic                 = len(coordinates.X_hub[0,:,0,0,0]) 
    phi_0                   = np.array([rotor.phase_offset_angle])  # phase angle offset  
    airfoils                = rotor.airfoils 
    num_sec                 = len(rotor.radius_distribution) 
    orientation             = np.array(rotor.orientation_euler_angles) * 1 
    body2thrust             = sp.spatial.transform.Rotation.from_rotvec(orientation).as_matrix() 
    commanded_thrust_vector =  np.atleast_2d(propulsor_conditions.commanded_thrust_vector_angle[cpt]) 
    for jj,airfoil in enumerate(airfoils):
        airfoil_points      = airfoil.number_of_points
    chord_coord             = int(np.floor(airfoil_points/2))
 
    # Lift and Drag - coefficients and distributions 
    fL      = aeroacoustic_data.disc_lift_distribution[cpt][None,:, :]
    fD      = aeroacoustic_data.disc_lift_distribution[cpt][None,:, :]
    CL      = aeroacoustic_data.disc_lift_coefficient[cpt][None,:, :]
    CD      = aeroacoustic_data.disc_drag_coefficient[cpt][None,:, :]
                
    y_u_6   = np.tile(aeroacoustic_data.blade_upper_surface[cpt][None, None, :, 0,None, None, :],(1,num_mic,1,num_h_b,num_h_l,1))
    y_l_6   = np.tile(aeroacoustic_data.blade_lower_surface[cpt][None, None, :, 0,None, None, :],(1,num_mic,1,num_h_b,num_h_l,1))
    
    # DFT to get loading modes
    CL_k           = sp.fft.rfft(CL, axis=2)
    CD_k           = sp.fft.rfft(CD, axis=2)
    fL_k           = sp.fft.rfft(fL, axis=2)
    fD_k           = sp.fft.rfft(fD, axis=2) 
    
    # ----------------------------------------------------------------------------------
    # Rotational Noise - Loading Noise
    # ----------------------------------------------------------------------------------  
    # [control point, microphones, rotors, radial distribution, blade harmonics, load harmonics]  
    
    # freestream density and speed of sound
    rho_3          = np.tile(freestream.density[cpt,:,None],(1,num_mic,num_h_b))
    a_3            = np.tile(freestream.speed_of_sound[cpt,:,None],(1,num_mic,num_h_b))
    
    B              = rotor.number_of_blades
    
    # blade harmonics
    m_3            = np.tile(harmonics_blade[None,None,:],(num_cpt,num_mic,1))
    m_4            = np.tile(harmonics_blade[None,None,:,None],(num_cpt,num_mic,1,num_h_l))
    m_5            = np.tile(harmonics_blade[None,None,None,:,None],(num_cpt,num_mic,num_sec,1,num_h_l))
    m_6            = np.tile(harmonics_blade[None,None,None,:,None,None],(num_cpt,num_mic,num_sec,1,num_h_l,chord_coord))
                                                                                            
    # loading harmonics
    k_4            = np.tile(harmonics_load[None,None,None,:],(num_cpt,num_mic,num_h_b,1))
    k_5            = np.tile(harmonics_load[None,None,None,None,:],(num_cpt,num_mic,num_sec,num_h_b,1))
    k_6            = np.tile(harmonics_load[None,None,None,None,:,None],(num_cpt,num_mic,num_sec,num_h_b,1,chord_coord))
    
    # reference atmospheric pressure
    p_ref          = 2E-5
    
    # net angle of inclination of propeller axis wrt inertial axis
    alpha_4        = np.tile((angle_of_attack + np.arccos(body2thrust[0,0]))[:,:,None,None],(1,num_mic,num_h_b,num_h_l))
    alpha_5        = np.tile((angle_of_attack + np.arccos(body2thrust[0,0]))[:,:,None,None,None],(1,num_mic,num_sec,num_h_b,num_h_l))
    alpha_6        = np.tile((angle_of_attack + np.arccos(body2thrust[0,0]))[:,:,None,None,None,None],(1,num_mic,num_sec,num_h_b,num_h_l,chord_coord))
    
    # rotor angular speed
    omega_3        = np.tile(aeroacoustic_data.omega[cpt,:,None],(1,num_mic,num_h_b))   
    
    R              = rotor.radius_distribution
    
    # Non-dimensional radius distribution
    z_5            = np.tile((R/R[-1])[None,None,:,None,None],(num_cpt,num_mic,1,num_h_b,num_h_l))
    z_6            = np.tile((R/R[-1])[None,None,:,None,None,None],(num_cpt,num_mic,1,num_h_b,num_h_l,chord_coord))
    
    # Radial chord distribution
    c_5            = np.tile(rotor.chord_distribution[None,None,:,None,None],(num_cpt,num_mic,1,num_h_b,num_h_l))
    c_6            = np.tile(rotor.chord_distribution[None,None,:,None,None,None],(num_cpt,num_mic,1,num_h_b,num_h_l,chord_coord))
    
    MCA_5          = np.tile(rotor.mid_chord_alignment[None,None,:,None,None],(num_cpt,num_mic,1,num_h_b,num_h_l))
    
    # chord to diameter ratio
    R_tip          = rotor.tip_radius
    D              = 2*R[-1]
    B_D_5          = c_5/D
    B_D_6          = c_6/D
    
    # maximum thickness to chord ratio
    t_b            = rotor.thickness_to_chord
    t_b_5          = np.tile(t_b[None,None,:,None,None],(num_cpt,num_mic,1,num_h_b,num_h_l))
    
    # chordwise thickness distribution normalized wrt chord
    H_6            = (y_u_6 - y_l_6)/c_6
    
    
    # Rotorcraft speed and mach number
    V_3            = np.tile(np.linalg.norm(velocity_vector, axis=1) [:,None,None],(1,num_mic,num_h_b))
    M_3            = V_3/a_3
    M_5            = np.tile(M_3[:,:,None,:,None],(1,1,num_sec,1,num_h_l))
    M_6            = np.tile(M_3[:,:,None,:,None,None],(1,1,num_sec,1,num_h_l,chord_coord))
    
    # Rotor tip speed and mach number
    V_tip          = R_tip*omega_3                                                        
    M_t_3          = V_tip/a_3
    M_t_5          = np.tile(M_t_3[:,:,None,:,None],(1,1,num_sec,1,num_h_l))
    M_t_6          = np.tile(M_t_3[:,:,None,:,None,None],(1,1,num_sec,1,num_h_l,chord_coord))
    
    # Section relative mach number
    M_r_5          = np.sqrt(M_5**2 + (z_5**2)*(M_t_5**2))
    
    # retarded theta
    theta_r        = coordinates.theta_hub_r[cpt,:,0,0]
    theta_r_3      = np.tile(theta_r[None,:,None],(1,1,num_h_b))
    theta_r_4      = np.tile(theta_r[None,:,None,None],(1,1,num_h_b,num_h_l))
    theta_r_5      = np.tile(theta_r[None,:,None,None,None],(1,1,num_sec,num_h_b,num_h_l))
    theta_r_6      = np.tile(theta_r[None,:,None,None,None,None],(1,1,num_sec,num_h_b,num_h_l,chord_coord))
    
    # retarded distance to source
    Y              = np.sqrt(coordinates.X_hub[cpt,:,0,0,1]**2 +  coordinates.X_hub[cpt,:,0,0,2] **2)
    Y_3            = np.tile(Y[None,:,None],(1,1,num_h_b))
    r_3            = Y_3/np.sin(theta_r_3)
    
    # phase angles
    phi_0_vec      = np.tile(phi_0[:,None,None,None],(num_cpt,num_mic,num_h_b,num_h_l))
    phi_4          = np.tile(coordinates.phi_hub_r[cpt,:,0,0,None,None],(1,1,num_h_b,num_h_l)) + phi_0_vec
    phi_5          = np.tile(phi_4[:,:,None,:,:],(1,1,num_sec,1,1))
    phi_6          = np.tile(phi_4[:,:,None,:,:,None],(1,1,num_sec,1,1,chord_coord))
    
    # total angle between propeller axis and r vector
    theta_r_prime_4 = np.arccos(np.cos(theta_r_4)*np.cos(alpha_4) + np.sin(theta_r_4)*np.sin(phi_4)*np.sin(alpha_4))
    theta_r_prime_5 = np.arccos(np.cos(theta_r_5)*np.cos(alpha_5) + np.sin(theta_r_5)*np.sin(phi_5)*np.sin(alpha_5))
    theta_r_prime_6 = np.arccos(np.cos(theta_r_6)*np.cos(alpha_6) + np.sin(theta_r_6)*np.sin(phi_6)*np.sin(alpha_6))
        
    phi_prime_4    = np.arccos((np.sin(theta_r_4)*np.cos(phi_4))/np.sin(theta_r_prime_4))
    
    # Velocity in the rotor frame
    T_body2inertial = conditions.frames.body.transform_to_inertial
    T_inertial2body = orientation_transpose(T_body2inertial)
    V_body          = orientation_product(T_inertial2body,velocity_vector)
    body2thrust,_   = rotor.body_to_prop_vel(commanded_thrust_vector)
    T_body2thrust   = orientation_transpose(body2thrust)
    V_thrust        = orientation_product(T_body2thrust,V_body)
    V_thrust_perp   = np.atleast_2d(V_thrust[cpt,0,None])
    V_thrust_perp_3 = np.tile(V_thrust_perp[:,:,None],(1,num_mic,num_h_b))
    M_thrust_3      = V_thrust_perp_3/a_3
    M_thrust_5      = np.tile(M_thrust_3[:,:,None,:,None],(1,1,num_sec,1,num_h_l))
    
    # helicoid angle
    zeta_5          = np.arctan(M_thrust_5/(z_5*M_t_5))
    zeta_6          = np.tile(zeta_5[:,:,:,:,:,None],(1,1,1,1,1,chord_coord))
    
    # wavenumbers
    k_m_3          = m_3*B*omega_3/a_3
    k_m_bar        = k_m_3/(1 - M_3*np.cos(theta_r_3))
    k_x_hat_5      = 2*B_D_5*(((m_5*B-k_5)*np.cos(zeta_5))/z_5 + (m_5*B*M_t_5*np.cos(theta_r_prime_5)*np.sin(zeta_5))/(1-M_5*np.cos(theta_r_5)))
    k_x_hat_6      = 2*B_D_6*(((m_6*B-k_6)*np.cos(zeta_6))/z_6 + (m_6*B*M_t_6*np.cos(theta_r_prime_6)*np.sin(zeta_6))/(1-M_6*np.cos(theta_r_6)))
    k_y_hat_5      = 2*B_D_5*(((m_5*B-k_5)*np.sin(zeta_5))/z_5 - (m_5*B*M_t_5*np.cos(theta_r_prime_5)*np.cos(zeta_5))/(1-M_5*np.cos(theta_r_5)))
    
    # phase angles
    phi_s_5        = k_x_hat_5*MCA_5/c_5
    # phi_FA         = k_x_hat*
    
    Noise.f          = B*m_3*omega_3/(2*np.pi)

    
    CL_k_5         = np.tile(CL_k[:,None,:,None,0:num_h_l],(1,num_mic,1,num_h_b,1))
    CD_k_5         = np.tile(CD_k[:,None,:,None,0:num_h_l],(1,num_mic,1,num_h_b,1))
    
    # [control point, microphones, rotors, radial distribution, blade harmonics, load harmonics, chordwise coordinate]
    fL_k_6         = np.tile(fL_k[:,None,:,None,0:num_h_l,:],(1,num_mic,1,num_h_b,1,1))
    fD_k_6         = np.tile(fD_k[:,None,:,None,0:num_h_l,:],(1,num_mic,1,num_h_b,1,1))
    
    
    # frequency domain source function for drag and lift
    X_edge         = np.linspace(-0.5,0.5,chord_coord+1)
    dX             = np.diff(X_edge)
    dX_tiled_6     = np.tile(dX[None,None,None,None,None,:],(num_cpt,num_mic,num_sec,num_h_b,num_h_l,1))
    X              = 0.5*(X_edge[0:-1] + X_edge[1:])
    X_6            = np.tile(X[None,None,None,None,None,:],(num_cpt,num_mic,num_sec,num_h_b,num_h_l,1))
    exp_term_6     = np.exp(1j*k_x_hat_6*X_6)
    psi_Lk_5       = np.trapz(fL_k_6*exp_term_6, x=X, axis=5)
    psi_Dk_5       = np.trapz(fD_k_6*exp_term_6, x=X, axis=5)
    
    psi_hat_Lk_5   = psi_Lk_5*np.exp(1j*(phi_s_5 + phi_5))
    psi_hat_Dk_5   = psi_Dk_5*np.exp(1j*(phi_s_5 + phi_5))
    psi_hat_Fk_5   = 0.5*(k_y_hat_5*CL_k_5*psi_hat_Lk_5 + k_x_hat_5*CD_k_5*psi_hat_Dk_5)
    
    
    # FREQUENCY DOMAIN PRESSURE TERM FOR LOADING
    J_mBk_5        = jv(m_5*B-k_5, (m_5*B*z_5*M_t_5*np.sin(theta_r_prime_5))/(1-M_5*np.cos(theta_r_5)))
    L_Integrand_5  = (M_r_5**2)*psi_hat_Fk_5*J_mBk_5
    L_Summand_4    = np.trapz(L_Integrand_5, x=z_5[0,0,:,0,0], axis=2)*np.exp(1j*(m_4*B-k_4)*(phi_prime_4-(np.pi/2)))
    L_Summation_3  = np.sum(L_Summand_4, axis=3)
    P_Lm           = (-1j*rho_3*(a_3**2)*B*np.exp(1j*k_m_3*r_3)*L_Summation_3)/(4*np.pi*(r_3/R_tip)*(1-M_3*np.cos(theta_r_3)))
    
    # frequency domain source function for drag and lift
    psi_V_5        = np.trapz(H_6*exp_term_6, x=X, axis=5)
    
    # FREQUENCY DOMAIN PRESSURE TERM FOR THICKNESS
    V_Integrand_5  = (M_r_5**2)*(k_x_hat_5**2)*t_b_5*psi_V_5*J_mBk_5
    V_Summand_4    = np.trapz(V_Integrand_5, x=z_5[0,0,:,0,0], axis=2)*np.exp(1j*m_4*B*(phi_prime_4-(np.pi/2)))
    
    # we take a single dimension along the 4th axis because we only want the loading mode corresponding to k=0
    V_Summation_3  = V_Summand_4[:,:,:,0]
    P_Vm           = (-rho_3*(a_3**2)*B*np.exp(1j*k_m_3*r_3)*V_Summation_3)/(4*np.pi*(r_3/R_tip)*(1-M_3*np.cos(theta_r_3)))
    
    
    # SOUND PRESSURE LEVELS
    P_Lm_abs       = np.abs(P_Lm)
    P_Vm_abs       = np.abs(P_Vm)
    Noise.SPL_prop_harmonic_bpf_spectrum     = 20*np.log10((abs(P_Lm_abs + P_Vm_abs))/p_ref)  
    Noise.SPL_prop_harmonic_1_3_spectrum     = convert_to_third_octave_band(Noise.SPL_prop_harmonic_bpf_spectrum,Noise.f,settings)          
    Noise.SPL_prop_harmonic_1_3_spectrum[np.isinf(Noise.SPL_prop_harmonic_1_3_spectrum)]         = 0 
    
    return
    


# RCAIDE/Framework/Analyses/Mission/Segments/Conditions/Numerics.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Conditions import Conditions 
from RCAIDE.Library.Methods.Utilities.Chebyshev  import chebyshev_data 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Numerics
# ----------------------------------------------------------------------------------------------------------------------

class Numerics(Conditions):
    """ Creates the data structure for the numerical solving of a mission.
    
        Assumptions:
        None
        
        Source:
        None
    """
    
    def __defaults__(self):
        """This sets the default values.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        self.tag                              = 'numerics' 
        self.number_of_control_points         = 16
        self.discretization_method            = chebyshev_data 
        self.solver_jacobian                  = "none"
        self.tolerance_solution               = 1e-8
        self.converged                        = None
        self.max_evaluations                  = 0.
        self.step_size                        = None
        
        self.dimensionless                    = Conditions()
        self.dimensionless.control_points     = np.empty([0,0])
        self.dimensionless.differentiate      = np.empty([0,0])
        self.dimensionless.integrate          = np.empty([0,0]) 
            
        self.time                             = Conditions()
        self.time.control_points              = np.empty([0,0])
        self.time.differentiate               = np.empty([0,0])
        self.time.integrate                   = np.empty([0,0]) 
        
        
        
        


# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/build_VLM_surrogates.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# package imports 
from scipy.interpolate                                           import RegularGridInterpolator 

# ----------------------------------------------------------------------------------------------------------------------
#  build_AVL_surrogates
# ---------------------------------------------------------------------------------------------------------------------- 
def build_AVL_surrogates(aerodynamics):
    """Build a surrogate using sample evaluation results.
    
    Assumptions:
        None
        
    Source:
        None

    Args:
        aerodynamics       : VLM analysis          [unitless] 
        
    Returns: 
        None  
    """
    surrogates  = aerodynamics.surrogates
    training    = aerodynamics.training  
    AoA_data    = training.angle_of_attack
    mach_data   = training.Mach
    

    CL_data       = training.coefficients[0,:,:]
    CDi_data      = training.coefficients[1,:,:]
    e_data        = training.coefficients[2,:,:]  
    CM_data       = training.coefficients[3,:,:]
    Cm_alpha_data = training.coefficients[4,:,:]
    Cn_beta_data  = training.coefficients[5,:,:]
    NP_data       = training.coefficients[6,:,:] 
   
    surrogates.lift_coefficient            = RegularGridInterpolator((AoA_data, mach_data), CL_data      ,method = 'linear',   bounds_error=False, fill_value=None)   
    surrogates.drag_coefficient            = RegularGridInterpolator((AoA_data, mach_data), CDi_data     ,method = 'linear',   bounds_error=False, fill_value=None)   
    surrogates.span_efficiency_factor      = RegularGridInterpolator((AoA_data, mach_data), e_data       ,method = 'linear',   bounds_error=False, fill_value=None)   
    surrogates.moment_coefficient          = RegularGridInterpolator((AoA_data, mach_data), CM_data      ,method = 'linear',   bounds_error=False, fill_value=None)   
    surrogates.Cm_alpha_moment_coefficient = RegularGridInterpolator((AoA_data, mach_data), Cm_alpha_data,method = 'linear',   bounds_error=False, fill_value=None)    
    surrogates.Cn_beta_moment_coefficient  = RegularGridInterpolator((AoA_data, mach_data), Cn_beta_data ,method = 'linear',   bounds_error=False, fill_value=None)   
    surrogates.neutral_point               = RegularGridInterpolator((AoA_data, mach_data), NP_data      ,method = 'linear',   bounds_error=False, fill_value=None)     
        
    return 


## @ingroup Methods-Utilities-Chebyshev
# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np

# ----------------------------------------------------------------------
#  Method
# ----------------------------------------------------------------------

## @ingroup Methods-Utilities-Chebyshev
def linear_data(N = 16, integration = True, **options):
    """Calculates the differentiation and integration matricies
    using chebyshev's pseudospectral algorithm, based on linearly
    spaced samples in x.
    
    D and I are not symmetric
    get derivatives with df_dy = np.dot(D,f)
    get integral with    int_f = np.dot(I,f)
        where f is either a 1-d vector or 2-d column array
        
    A full example of how these operators are used is available in 
    the chebyshev_data.py (same folder)

    Assumptions:
    None

    Source:
    N/A

    Inputs:
    N                      [-]        Number of points
    integration (optional) <boolean>  Determines if the integration operator is calculated

    Outputs:
    x                      [-]        N-number of cosine spaced control points, in range [0,1]
    D                      [-]        Differentiation operation matrix
    I                      [-]        Integration operation matrix, or None if integration = False

    Properties Used:
    N/A
    """           
    
    # setup
    N = int(N)
    if N <= 0: raise RuntimeError("N = %i, must be > 0" % N)
    
    
    # --- X vector
    
    # linear spaced in range [0,1]
    x = np.linspace(0,1,N)   


    # --- Differentiation Operator
    
    # coefficients
    c = np.array( [2.] + [1.]*(N-2) + [2.] )
    c = c * ( (-1.) ** np.arange(0,N) )
    A = np.tile( x, (N,1) ).T
    dA = A - A.T + np.eye( N )
    cinv = 1./c; 

    # build operator
    D = np.zeros( (N,N) );
    
    # math
    c    = np.array(c)
    cinv = np.array([cinv])
    cs   = np.multiply(c,cinv.T)
    D    = np.divide(cs.T,dA)

    # more math
    D = D - np.diag( np.sum( D.T, axis=0 ) );

    # --- Integration operator
    
    if integration:
        # invert D except first row and column
        I = np.linalg.inv(D[1:,1:]); 
        
        # repack missing columns with zeros
        I = np.append(np.zeros((1,N-1)),I,axis=0)
        I = np.append(np.zeros((N,1)),I,axis=1)
        
    else:
        I = None
        
    # done!
    return x, D, I


# RCAIDE/Library/Components/Propulsors/Converters/Prop_Rotor.py
# 
# 
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports 
from RCAIDE.Framework.Core import Data
from .Rotor      import Rotor

# ----------------------------------------------------------------------------------------------------------------------
#  PROP-ROTOR CLASS
# ----------------------------------------------------------------------------------------------------------------------           
class Prop_Rotor(Rotor):
    """
    A prop-rotor component model for tiltrotor and convertible aircraft applications, inheriting from the base Rotor class.

    Attributes
    ----------
    tag : str
        Identifier for the prop-rotor. Default is 'prop_rotor'.
        
    orientation_euler_angles : list
        Vector of angles [rad] defining rotor orientation [θ, φ, ψ]. 
        Default is [0., 0., 0.] for X-direction thrust in vehicle frame.
        
    use_2d_analysis : bool
        Flag for using 2D aerodynamic analysis. Default is False.
        
    variable_pitch : bool
        Flag indicating if rotor has variable pitch capability. Default is True.
        
    hover : Data
        Hover mode performance parameters
        
        - design_thrust : float
            Design point thrust [N]. Default is None.
        - design_torque : float
            Design point torque [N·m]. Default is None.
        - design_power : float
            Design point power [W]. Default is None.
        - design_angular_velocity : float
            Design point rotational speed [rad/s]. Default is None.
        - design_tip_mach : float
            Design point blade tip Mach number. Default is None.
        - design_acoustics : Data
            Acoustic characteristics at design point. Default is None.
        - design_performance : Data
            Performance metrics at design point. Default is None.
        - design_freestream_velocity : float
            Design point forward velocity [m/s]. Default is None.
        - design_SPL_dBA : float
            Design point sound pressure level [dBA]. Default is None.
        - design_pitch_command : float
            Design point blade pitch command [rad]. Default is 0.0.
        - design_Cl : float
            Design point lift coefficient. Default is None.
        - design_thrust_coefficient : float
            Design point thrust coefficient. Default is None.
        - design_power_coefficient : float
            Design point power coefficient. Default is None.
            
    cruise : Data
        Cruise mode performance parameters
        
        - design_thrust : float
            Design point thrust [N]. Default is None.
        - design_torque : float
            Design point torque [N·m]. Default is None.
        - design_power : float
            Design point power [W]. Default is None.
        - design_angular_velocity : float
            Design point rotational speed [rad/s]. Default is None.
        - design_tip_mach : float
            Design point blade tip Mach number. Default is None.
        - design_acoustics : Data
            Acoustic characteristics at design point. Default is None.
        - design_performance : Data
            Performance metrics at design point. Default is None.
        - design_SPL_dBA : float
            Design point sound pressure level [dBA]. Default is None.
        - design_pitch_command : float
            Design point blade pitch command [rad]. Default is 0.0.
        - design_Cl : float
            Design point lift coefficient. Default is None.
        - design_thrust_coefficient : float
            Design point thrust coefficient. Default is None.
        - design_power_coefficient : float
            Design point power coefficient. Default is None.
            
    oei : Data
        One engine inoperative performance parameters
        (Similar structure to hover and cruise attributes)

    optimization_parameters : Data
        - multiobjective_performance_weight : float
            Weight factor for multi-objective optimization. Default is 0.5.

    Notes
    -----
    The Prop_Rotor class models rotors designed for both hover and forward flight,
    typical of tiltrotor aircraft. It includes capabilities for:

    * Hover and cruise performance analysis
    * One engine inoperative (OEI) conditions
    * Variable pitch operation
    * Acoustic analysis
    * Performance optimization
    * Multi-mode operation

    **Definitions**
    
    'OEI'
        One Engine Inoperative - emergency condition where one engine fails
    'SPL'
        Sound Pressure Level - measure of acoustic intensity
    'Tip Mach'
        Mach number at the blade tip, including rotational and forward flight effects
    'Multi-objective Performance'
        Combined optimization of multiple performance metrics

    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Converters.Rotor
    RCAIDE.Library.Components.Propulsors.Converters.Lift_Rotor
    """     
    def __defaults__(self):
        """This sets the default values for the component to function.
        
        Assumptions:
        None
        
        Source:
        N/A
        
        Inputs:
        None
        
        Outputs:
        None
        
        Properties Used:
        None
        """         

        self.tag                                 = 'prop_rotor'
        self.orientation_euler_angles            = [0.,0.,0.] # This is X-direction thrust in vehicle frame
        self.use_2d_analysis                     = False       
        self.variable_pitch                      = True 
        
        self.hover                               = Data()
        self.hover.design_thrust                 = None
        self.hover.design_torque                 = None
        self.hover.design_power                  = None
        self.hover.design_angular_velocity       = None
        self.hover.design_tip_mach               = None
        self.hover.design_acoustics              = None
        self.hover.design_performance            = None
        self.hover.design_freestream_velocity    = None
        self.hover.design_SPL_dBA                = None
        self.hover.design_pitch_command          = 0.0
        self.hover.design_Cl                     = None
        self.hover.design_thrust_coefficient     = None
        self.hover.design_power_coefficient      = None  
        
        self.oei                                 = Data()   
        self.oei.design_thrust                   = None
        self.oei.design_torque                   = None
        self.oei.design_power                    = None
        self.oei.design_angular_velocity         = None
        self.oei.design_tip_mach                 = None
        self.oei.design_acoustics                = None
        self.oei.design_performance              = None 
        self.oei.design_freestream_velocity      = None   
        self.oei.design_pitch_command            = 0.0
        self.oei.design_altitude                 = None
        self.oei.design_SPL_dBA                  = None
        self.oei.design_Cl                       = None
        self.oei.design_thrust_coefficient       = None
        self.oei.design_power_coefficient        = None  

        self.cruise                              = Data()     
        self.cruise.design_thrust                = None
        self.cruise.design_torque                = None
        self.cruise.design_power                 = None
        self.cruise.design_angular_velocity      = None
        self.cruise.design_tip_mach              = None
        self.cruise.design_acoustics             = None
        self.cruise.design_performance           = None
        self.cruise.design_SPL_dBA               = None
        self.cruise.design_pitch_command         = 0.0
        self.cruise.design_Cl                    = None
        self.cruise.design_thrust_coefficient    = None
        self.cruise.design_power_coefficient     = None       
        
        self.optimization_parameters.multiobjective_performance_weight  = 0.5



## @ingroup Methods-Noise-Frequency_Domain_Buildup-Rotor
# RCAIDE/Methods/Noise/Frequency_Domain_Buildup/Rotor/compute_rotor_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports 
from RCAIDE.Framework.Core import  Data   
from RCAIDE.Library.Methods.Noise.Common.decibel_arithmetic                                import SPL_arithmetic  
from RCAIDE.Library.Methods.Noise.Common.compute_noise_source_coordinates                  import compute_rotor_point_source_coordinates  
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.harmonic_noise_point      import harmonic_noise_point
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.harmonic_noise_line       import harmonic_noise_line
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.harmonic_noise_plane      import harmonic_noise_plane 
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.broadband_noise           import broadband_noise
from RCAIDE.Library.Methods.Noise.Common                                                   import atmospheric_attenuation
from RCAIDE.Library.Methods.Noise.Metrics.A_weighting_metric                               import A_weighting_metric  
from RCAIDE.Library.Methods.Geometry.Airfoil.import_airfoil_geometry                       import import_airfoil_geometry
from RCAIDE.Library.Methods.Aerodynamics.Airfoil_Panel_Method.airfoil_analysis             import airfoil_analysis

# Python package imports   
import numpy as np    
from RCAIDE.Framework.Core import interp2d 

# ----------------------------------------------------------------------------------------------------------------------    
#  Rotor Noise 
# ----------------------------------------------------------------------------------------------------------------------    
## @ingroup Methods-Noise-Frequency_Domain_Buildup-Rotor
def compute_rotor_noise(microphone_locations,propulsor,rotor,segment,settings, rotor_index = 0, previous_rotor_tag = None, identical_propulsors=True):
    ''' This is a collection medium-fidelity frequency domain methods for rotor acoustic noise prediction which 
    computes the acoustic signature (sound pressure level, weighted sound pressure levels,
    and frequency spectrums of a system of rotating blades           
        
    Assumptions:
    None

    Source:
    None
    
    Inputs:
        rotors                  - data structure of rotors                            [None]
        segment                 - flight segment data structure                       [None] 
        results                 - data structure containing of acoustic data          [None]
        settings                - accoustic settings                                  [None]
                               
    Outputs:
        Results.    
            blade_passing_frequencies      - blade passing frequencies                           [Hz]
            SPL                            - total SPL                                           [dB]
            SPL_dBA                        - dbA-Weighted SPL                                    [dBA]
            SPL_1_3_spectrum               - 1/3 octave band spectrum of SPL                     [dB]
            SPL_1_3_spectrum_dBA           - 1/3 octave band spectrum of A-weighted SPL          [dBA]
            SPL_broadband_1_3_spectrum     - 1/3 octave band broadband contribution to total SPL [dB] 
            SPL_harmonic_1_3_spectrum      - 1/3 octave band harmonic contribution to total SPL  [dB]
            SPL_harmonic_bpf_spectrum_dBA  - A-weighted blade passing freqency spectrum of 
                                             harmonic compoment of SPL                           [dB]
            SPL_harmonic_bpf_spectrum      - blade passing freqency spectrum of harmonic
                                             compoment of SPL                                    [dB] 
     
    Properties Used:
        N/A   
    '''
 
    # unpack 
    conditions           = segment.state.conditions
    propulsor_conditions = conditions.energy[propulsor.tag]
    harmonics_blade      = settings.harmonics
    harmonics_load       = np.linspace(0,5,6).astype(int)  
    num_mic              = len(microphone_locations[:,0]) 
    num_cpt              = conditions._size
    num_f                = len(settings.center_frequencies)
      
    # create data structures for computation
    Noise   = Data()  
    Results = Data()

    Results.SPL                                           = np.zeros((num_cpt,num_mic))
    Results.SPL_dBA                                       = np.zeros_like(Results.SPL)
    Results.SPL_harmonic                                  = np.zeros_like(Results.SPL)
    Results.SPL_broadband                                 = np.zeros_like(Results.SPL)
    Results.blade_passing_frequencies                     = np.zeros(num_f)
    Results.SPL_1_3_spectrum                              = np.zeros((num_cpt,num_mic,num_f)) 
    Results.SPL_harmonic_bpf_spectrum                     = np.zeros_like(Results.SPL_1_3_spectrum)
    Results.SPL_harmonic_bpf_spectrum_dBA                 = np.zeros_like(Results.SPL_1_3_spectrum)
    Results.one_third_frequency_spectrum                  = settings.center_frequencies 
    Results.SPL_1_3_spectrum                              = np.zeros_like(Results.SPL_1_3_spectrum)
    Results.SPL_1_3_spectrum_dBA                          = np.zeros_like(Results.SPL_1_3_spectrum)
    Results.SPL_harmonic_1_3_spectrum                     = np.zeros_like(Results.SPL_1_3_spectrum)
    Results.SPL_harmonic_1_3_spectrum_dBA                 = np.zeros_like(Results.SPL_1_3_spectrum)
    Results.SPL_broadband_1_3_spectrum                    = np.zeros_like(Results.SPL_1_3_spectrum)
    Results.SPL_broadband_1_3_spectrum_dBA                = np.zeros_like(Results.SPL_1_3_spectrum)

    # compute position vector from point source (or should it be origin) at rotor hub to microphones 
    coordinates   = compute_rotor_point_source_coordinates(propulsor,rotor,conditions,microphone_locations,settings)        

    for cpt in range(num_cpt): 
        # ----------------------------------------------------------------------------------
        # Harmonic Noise
        # ---------------------------------------------------------------------------------- 
        # harmonic noise with planar load distribution
        if settings.fidelity == 'plane_source': 
            aeroacoustic_data = propulsor_conditions[rotor.tag]       
            Re                = aeroacoustic_data.disc_reynolds_number
            AOA_sec           = aeroacoustic_data.disc_effective_angle_of_attack  
            a_loc             = rotor.airfoil_polar_stations
            num_az            = aeroacoustic_data.number_azimuthal_stations     
            airfoils          = rotor.airfoils         
            for jj,airfoil in enumerate(airfoils):
                airfoil_points      = airfoil.number_of_points 
            chord_coord             = int(np.floor(airfoil_points/2))       
                
            if (identical_propulsors == False) and rotor_index !=0: 
                prev_aeroacoustic_data                   = propulsor_conditions[previous_rotor_tag]                 
                prev_aeroacoustic_data                   = propulsor_conditions[rotor.tag]  
                aeroacoustic_data.disc_lift_distribution = prev_aeroacoustic_data.disc_lift_distribution
                aeroacoustic_data.disc_drag_distribution = prev_aeroacoustic_data.disc_lift_distribution
                aeroacoustic_data.disc_lift_coefficient  = prev_aeroacoustic_data.disc_lift_coefficient 
                aeroacoustic_data.disc_drag_coefficient  = prev_aeroacoustic_data.disc_drag_coefficient  
                aeroacoustic_data.blade_upper_surface    = prev_aeroacoustic_data.blade_upper_surface
                aeroacoustic_data.blade_lower_surface    = prev_aeroacoustic_data.blade_lower_surface
            else: 
                # Lift and Drag - coefficients and distributions 
                fL      = np.tile(np.zeros_like(Re)[:,:,:,None],(1,1,1,chord_coord))
                fD      = np.zeros_like(fL)
                CL      = np.zeros_like(Re)
                CD      = np.zeros_like(Re) 
                y_up    = np.zeros_like(fL)
                y_low   = np.zeros_like(fL)
                                  
                for jj,airfoil in enumerate(airfoils):    
                    locs                  = np.where(np.array(a_loc) == jj ) 
                    alpha_azi             = np.atleast_2d(AOA_sec[cpt,locs,:].flatten())
                    Re_azi                = np.atleast_2d(Re[cpt,locs,:].flatten())      
                    pd                    = airfoil.polars 
                    if settings.use_plane_loading_surrogate: 
                        fL[cpt,locs,:,:]      = pd.lift_distribution_func((alpha_azi,Re_azi)).reshape(1,len(a_loc), num_az,chord_coord) 
                        fD[cpt,locs,:,:]      = pd.drag_distribution_func((alpha_azi,Re_azi)).reshape(1,len(a_loc), num_az,chord_coord)  
                        cl_invisc             = interp2d(Re_azi,alpha_azi,pd.reynolds_numbers, pd.angle_of_attacks, pd.lift_coefficients)
                        cd_visc               = interp2d(Re_azi,alpha_azi,pd.reynolds_numbers, pd.angle_of_attacks, pd.drag_coefficients)   
                        CL[cpt,locs,:]        = cl_invisc.reshape(1, len(a_loc), num_az) 
                        CD[cpt,locs,:]        = cd_visc.reshape(1, len(a_loc), num_az)                             
                    
                    else : 
                        airfoil_geometry      = import_airfoil_geometry(airfoil.coordinate_file,airfoil_points)
                        airfoil_properties    = airfoil_analysis(airfoil_geometry,alpha_azi,Re_azi)
                        fL[cpt,locs,:,:]      = airfoil_properties.fL.reshape(chord_coord, len(a_loc), num_az,1).swapaxes(0, 3)
                        fD[cpt,locs,:,:]      = airfoil_properties.fD.reshape(chord_coord, len(a_loc), num_az,1).swapaxes(0, 3)
                        CL[cpt,locs,:]        = airfoil_properties.cl_invisc.reshape(1, len(a_loc), num_az) 
                        CD[cpt,locs,:]        = airfoil_properties.cd_visc.reshape(1, len(a_loc), num_az) 
                        
                    y_up[cpt,locs,:,:]    = airfoil.geometry.y_upper_surface
                    y_low[cpt,locs,:,:]   = airfoil.geometry.y_lower_surface
                        
                aeroacoustic_data.disc_lift_distribution = fL
                aeroacoustic_data.disc_drag_distribution = fD
                aeroacoustic_data.disc_lift_coefficient  = CL
                aeroacoustic_data.disc_drag_coefficient  = CD 
                aeroacoustic_data.blade_upper_surface    = y_up
                aeroacoustic_data.blade_lower_surface    = y_low                        
                        
            harmonic_noise_plane(harmonics_blade,harmonics_load,conditions,propulsor_conditions,coordinates,rotor,settings,Noise,cpt)
        elif settings.fidelity == 'line_source': 
            harmonic_noise_line(harmonics_blade,harmonics_load,conditions,propulsor_conditions,coordinates,rotor,settings,Noise,cpt)
        else:
            harmonic_noise_point(harmonics_blade,harmonics_load,conditions,propulsor_conditions,coordinates,rotor,settings,Noise,cpt) 
    
        # ----------------------------------------------------------------------------------    
        # Broadband Noise
        # ---------------------------------------------------------------------------------- 
        broadband_noise(conditions,propulsor_conditions,coordinates,rotor,settings,Noise,cpt)  
    
        # ----------------------------------------------------------------------------------    
        # Atmospheric attenuation 
        # ----------------------------------------------------------------------------------
        delta_atmo = atmospheric_attenuation(np.linalg.norm(coordinates.X_r[:,0,0,0,:],axis=1),settings.center_frequencies)
    
        # ----------------------------------------------------------------------------------    
        # Combine Harmonic (periodic/tonal) and Broadband Noise
        # ----------------------------------------------------------------------------------
        num_mic      = len(coordinates.X_hub[0,:,0,0])
        SPL_total_1_3_spectrum      = 10*np.log10( 10**(Noise.SPL_prop_harmonic_1_3_spectrum/10) + 10**(Noise.SPL_prop_broadband_1_3_spectrum/10)) - np.tile(delta_atmo[cpt,None,:],(1,num_mic,1))  
        SPL_total_1_3_spectrum[np.isnan(SPL_total_1_3_spectrum)] = 0 
    
        # ----------------------------------------------------------------------------------
        # Summation of spectra from propellers into one SPL and store results
        # ----------------------------------------------------------------------------------
        Results.SPL[cpt,:]                                 = SPL_arithmetic(SPL_total_1_3_spectrum[0], sum_axis=1) 
        Results.SPL_dBA[cpt,:]                             = SPL_arithmetic(A_weighting_metric(SPL_total_1_3_spectrum[0],settings.center_frequencies), sum_axis=1) 
        Results.SPL_harmonic[cpt,:]                        = SPL_arithmetic(Noise.SPL_prop_harmonic_1_3_spectrum[0], sum_axis=1)
        Results.SPL_broadband[cpt,:]                       = SPL_arithmetic(Noise.SPL_prop_broadband_1_3_spectrum[0], sum_axis=1) 
          
        # blade passing frequency   
        Results.blade_passing_frequencies                  = Noise.f          
        Results.SPL_harmonic_bpf_spectrum[cpt,:,:]         = Noise.SPL_prop_harmonic_bpf_spectrum 
        Results.SPL_harmonic_bpf_spectrum_dBA[cpt,:,:]     = A_weighting_metric(Results.SPL_harmonic_bpf_spectrum[cpt,:,:],Noise.f) 
          
        # 1/3 octave band   
        Results.SPL_1_3_spectrum[cpt,:,:]                  = SPL_total_1_3_spectrum 
        Results.SPL_1_3_spectrum_dBA[cpt,:,:]              = A_weighting_metric(Results.SPL_1_3_spectrum[cpt,:,:],settings.center_frequencies)      
        Results.SPL_harmonic_1_3_spectrum[cpt,:,:]         = Noise.SPL_prop_harmonic_1_3_spectrum 
        Results.SPL_harmonic_1_3_spectrum_dBA[cpt,:,:]     = A_weighting_metric(Results.SPL_harmonic_1_3_spectrum[cpt,:,:],settings.center_frequencies) 
        Results.SPL_broadband_1_3_spectrum[cpt,:,:]        = Noise.SPL_prop_broadband_1_3_spectrum 
        Results.SPL_broadband_1_3_spectrum_dBA[cpt,:,:]    = A_weighting_metric(Results.SPL_broadband_1_3_spectrum[cpt,:,:],settings.center_frequencies) 
    
    # A-weighted
    conditions.noise[propulsor.tag][rotor.tag] = Results 
    return rotor.tag 



# RCAIDE/Framework/Analyses/Emissions/Emissions.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from RCAIDE.Framework.Core import Data
from RCAIDE.Framework.Analyses import Analysis
import  numpy as  np

# ----------------------------------------------------------------------
#  Emissions
# ---------------------------------------------------------------------- 
class Emissions(Analysis):
    """ The Top Level Emissions Analysis Class 
    """
    def __defaults__(self):
        """This sets the default values and methods for the analysis.
    
            Assumptions:
            None
    
            Source:
            None 
            """                   
        self.tag      = 'Emissions'    
        self.vehicle  = Data()
        self.settings = Data() 
    
        
    def evaluate(self,state):
        """The default evaluate function.

        Assumptions:
        None

        Source:
        None 

        Inputs:
        self   - emissions analyses 
        state  - flight conditions 

        Outputs:
        results  
        """         
        results = Data()
        
        return results 


# RCAIDE/Compoments/Nacelles/Body_of_Revolution_Nacelle.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core     import Data  
from .Nacelle                  import Nacelle 
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Body_of_Revolution_Nacelle
# ----------------------------------------------------------------------------------------------------------------------  
class Body_of_Revolution_Nacelle(Nacelle):
    """
    A nacelle design generated by rotating an airfoil profile around a central axis, 
    commonly used for podded engine installations.

    Attributes
    ----------
    tag : str
        Unique identifier for the nacelle component, defaults to 'body_of_revolution_nacelle'
        
    Airfoil : Data
        Container for airfoil profile data used to generate the nacelle shape, 
        defaults to empty Data()

    Notes
    -----
    A body of revolution nacelle is created by rotating an airfoil section about a 
    central axis, creating an axisymmetric shape. This approach is commonly used for:
    
    * Podded engine installations
    * Streamlined auxiliary power unit enclosures
    * External fuel tank designs
    
    **Major Assumptions**
    
    * Axisymmetric geometry
    * Smooth surface transitions
    * No significant flow separation
    
    **Definitions**

    'Body of Revolution'
        A three-dimensional shape created by rotating a two-dimensional profile 
        around a central axis
        
    'Highlight Radius'
        The radius at the forward-most point of the nacelle inlet

    See Also
    --------
    RCAIDE.Library.Components.Nacelles.Nacelle
        Base nacelle class
    RCAIDE.Library.Components.Nacelles.Stack_Nacelle
        Alternative nacelle design approach
    """
    
    def __defaults__(self):
        """
        Sets default values for the body of revolution nacelle attributes.
        """      
        self.tag     = 'body_of_revolution_nacelle' 
        self.Airfoil = None
    
    def append_airfoil(self, airfoil):
        """
        Adds an airfoil profile to the nacelle definition.

        Parameters
        ----------
        airfoil : Data
            Airfoil profile data to be used in generating the nacelle shape
        """
        # Assert database type
        if not isinstance(airfoil,RCAIDE.Library.Components.Airfoils.Airfoil):
            raise Exception('input component must be of type Airfoil')

        # Store data
        self.airfoil = airfoil

        return            



# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
from   RCAIDE import  * 
from   RCAIDE.Framework.Core import Units
import numpy as np

# ----------------------------------------------------------------------
#  compute_slat_lift
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Fidelity_Zero-Lift
def compute_slat_lift(slat_angle,sweep_angle):
    """Computes the increase in lift due to slats

    Assumptions:
    None

    Source:
    adg.stanford.edu (Stanford AA241 A/B Course Notes)

    Inputs:
    slat_angle   [radians]
    sweep_angle  [radians]

    Outputs:
    dcl_slat     [Unitless]

    Properties Used:
    N/A
    """     

    # unpack
    sa = slat_angle  / Units.deg
    sw = sweep_angle

    # AA241 Method from adg.stanford.edu
    dcl_slat = (sa/23.)*(np.cos(sw))**1.4 * np.cos(sa * Units.deg)**2

    #returning dcl_slat
    return dcl_slat
 


# RCAIDE/Framework/Analyses/Mission/Segments/Descent/Constant_EAS_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# RCAIDE imports
from RCAIDE.Framework.Core                                 import Units 
from RCAIDE.Framework.Mission.Segments.Evaluate   import Evaluate 
from RCAIDE.Library.Mission                      import Common,Segments

# ----------------------------------------------------------------------------------------------------------------------  
#  Constant_EAS_Constant_Rate
# ----------------------------------------------------------------------------------------------------------------------  
class Constant_EAS_Constant_Rate(Evaluate):
    """ Fixed at an Equivalent Airspeed (EAS) the vehicle will descent at a constant rate.
    
        Assumptions:
        None
        
        Source:
        None
    """        
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """            
        
        # -------------------------------------------------------------------------------------------------------------- 
        #   User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
        self.altitude_start       = None # Optional
        self.altitude_end         = 10. * Units.km
        self.descent_rate         = 3.  * Units.m / Units.s
        self.equivalent_air_speed = 100 * Units.m / Units.s
        self.true_course          = 0.0 * Units.degrees      
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # -------------------------------------------------------------------------------------------------------------- 
        initialize                         = self.process.initialize  
        initialize.differentials_altitude  = Common.Initialize.differentials_altitude
        initialize.conditions              = Segments.Descent.Constant_EAS_Constant_Rate.initialize_conditions  
        iterate                            = self.process.iterate
        iterate.unknowns.mission           = Common.Unpack_Unknowns.orientation   
        iterate.unknowns.controls          = Common.Unpack_Unknowns.control_surfaces
        iterate.residuals.flight_dynamics  = Common.Residuals.flight_dynamics

        return
       


# RCAIDE/Library/Methods/Emissions/Chemical_Reactor_Network_Method/evaluate_CRN_emission_indices.py
#  
# Created: Jul 2024, M. Clarke, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import  RCAIDE
from    RCAIDE.Framework.Core import Data
from    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera import evaluate_cantera 
 
# package imports
import  numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  evaluate_correlation_emissions_indices
# ---------------------------------------------------------------------------------------------------------------------- 
def evaluate_CRN_emission_indices_no_surrogate(segment,settings,vehicle):

    """
    Computes emission indices directly using Chemical Reactor Network without surrogate models.

    Parameters
    ----------
    segment : Data
        Mission segment data container
        
        - state : Data
            Current state 

            - numerics : Data
                Numerical integration parameters

                - time : Data
                    Time integration settings
            - conditions : Data
                Flight conditions and component states
            - ones_row : function
                Creates array of ones with specified size
                
    settings : Data
        Configuration settings for the simulation
        
    vehicle : Data
        Vehicle configuration data

        - networks : list
            List of propulsion system networks

            - fuel_lines : list
                Fuel distribution systems
            - propulsors : list
                Propulsion units 
    Returns
    -------
        Updates segment.state.conditions.emissions with:
        
        total : Data
            Total emissions over segment

            - CO2 : float
                Total CO2 emissions [kg]
            - H2O : float
                Total H2O emissions [kg]
            - NOx : float
                Total NOx emissions [kg]
        index : Data
            Emission indices

            - CO2 : ndarray
                CO2 emission index [kg_CO2/kg_fuel]
            - CO : ndarray
                CO emission index [kg_CO/kg_fuel]
            - H2O : ndarray
                H2O emission index [kg_H2O/kg_fuel]
            - NO : ndarray
                NO emission index [kg_NO/kg_fuel]
            - NO2 : ndarray
                NO2 emission index [kg_NO2/kg_fuel]

    Notes
    -----
    Computes emissions by directly evaluating the chemical kinetics at each time step
    using Cantera. 

    **Extra modules required**

    * numpy
    * Cantera (through evaluate_cantera function)

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera 

    References
    ----------
    [1] Goodwin, D. G., Speth, R. L., Moffat, H. K., & Weber, B. W. (2023). Cantera: An object-oriented software toolkit for chemical kinetics, thermodynamics, and transport processes (Version 3.0.0) [Computer software]. https://www.cantera.org
    """
  
    # unpack
    state     = segment.state
    I         = state.numerics.time.integrate
    
    CO2_total = 0 * state.ones_row(1)  
    CO_total  = 0 * state.ones_row(1) 
    H2O_total = 0 * state.ones_row(1) 
    NO_total  = 0 * state.ones_row(1) 
    NO2_total = 0 * state.ones_row(1) 

    for network in vehicle.networks:  
        for fuel_line in network.fuel_lines:
            if fuel_line.active:  
                for p_i ,  propulsor in enumerate(network.propulsors):
                    if propulsor.active == True: 
                        if (type(propulsor) == RCAIDE.Library.Components.Propulsors.Turbofan) or \
                            type(propulsor) == RCAIDE.Library.Components.Propulsors.Turboshaft or \
                            type(propulsor) == RCAIDE.Library.Components.Propulsors.Turboprop or \
                            type(propulsor) == RCAIDE.Library.Components.Propulsors.Turbojet:    
                        
                            combustor = propulsor.combustor
                        
                            # unpack component conditions
                            n_cp                 = state.numerics.number_of_control_points 
                            propulsor_conditions = state.conditions.energy[propulsor.tag] 
                            combustor_conditions = propulsor_conditions[combustor.tag]  

                            
                            T    = combustor_conditions.inputs.stagnation_temperature
                            P    = combustor_conditions.inputs.stagnation_pressure 
                            mdot = propulsor_conditions.core_mass_flow_rate 
                            FAR  = combustor_conditions.outputs.fuel_to_air_ratio 

                            EI_CO2_comb   = 0 * state.ones_row(1)   
                            EI_CO_comb    = 0 * state.ones_row(1)  
                            EI_H2O_comb   = 0 * state.ones_row(1)  
                            EI_NO_comb    = 0 * state.ones_row(1)  
                            EI_NO2_comb   = 0 * state.ones_row(1)                              
                            if network.identical_propulsors == True and p_i != 0:
                                EI_CO2_comb = EI_CO2_prev
                                EI_CO_comb  = EI_CO_prev
                                EI_H2O_comb = EI_H2O_prev
                                EI_NO_comb  = EI_NO_prev
                                EI_NO2_comb = EI_NO2_prev 
                                
                            else:     
                                for t_idx in range(n_cp):
                                    # Call cantera 
                                    results = evaluate_cantera(combustor,T[t_idx,0],P[t_idx,0],mdot[t_idx,0],FAR[t_idx,0])
                                    
                                    EI_CO2_comb[t_idx,0] = results.EI_CO2
                                    EI_CO_comb[t_idx,0]  = results.EI_CO 
                                    EI_H2O_comb[t_idx,0] = results.EI_H2O
                                    EI_NO_comb[t_idx,0]  = results.EI_NO 
                                    EI_NO2_comb[t_idx,0] = results.EI_NO2
                                    
                                    EI_CO2_prev = EI_CO2_comb 
                                    EI_CO_prev  =  EI_CO_comb  
                                    EI_H2O_prev = EI_H2O_comb 
                                    EI_NO_prev  =  EI_NO_comb  
                                    EI_NO2_prev = EI_NO2_comb 
                                
                            CO2_total  += np.dot(I,mdot*EI_CO2_comb)
                            CO_total   += np.dot(I,mdot *EI_CO_comb )
                            H2O_total  += np.dot(I,mdot*EI_H2O_comb)
                            NO_total   += np.dot(I,mdot *EI_NO_comb ) 
                            NO2_total  += np.dot(I,mdot *EI_NO2_comb)

    emissions                 = Data()
    emissions.total           = Data()
    emissions.index           = Data() 
    emissions.total.CO2       = CO2_total  * combustor.fuel_data.global_warming_potential_100.CO2 
    emissions.total.H2O       = H2O_total  * combustor.fuel_data.global_warming_potential_100.H2O  
    emissions.total.NOx       = (NO_total + NO2_total) * combustor.fuel_data.global_warming_potential_100.NOx 
    emissions.index.CO2       = EI_CO2_comb
    emissions.index.CO        = EI_CO_comb 
    emissions.index.H2O       = EI_H2O_comb
    emissions.index.NO        = EI_NO_comb 
    emissions.index.NO2       = EI_NO2_comb 
    
    state.conditions.emissions =  emissions
    return    
    

def evaluate_CRN_emission_indices_surrogate(segment,settings,vehicle): 

    """
    Computes emission indices using pre-trained Chemical Reactor Network surrogate models.

    Parameters
    ----------
    segment : Data
        Mission segment data container
        
        - state : Data
            Current state of the system

            - numerics : Data
                Numerical integration parameters
            - conditions : Data
                Flight conditions and component states
        - analyses : Data
            Analysis settings and models

            - emissions : Data
                Emissions analysis settings

                - surrogates : Data
                    Trained surrogate models for each species
                
    settings : Data
        Configuration settings for the simulation
        
    vehicle : Data
        Vehicle configuration data

        - networks : list
            List of propulsion system networks

            - propulsors : list
                Propulsion units 

    Returns
    -------
    Updates segment.state.conditions.emissions with:
        
    total : Data
        Total emissions over segment

        - CO2 : float
            Total CO2 emissions [kg]
        - H2O : float
            Total H2O emissions [kg]
        - NOx : float
            Total NOx emissions [kg]
    index : Data
        Emission indices

        - CO2 : ndarray
            CO2 emission index [kg_CO2/kg_fuel]
        - CO : ndarray
            CO emission index [kg_CO/kg_fuel]
        - H2O : ndarray
            H2O emission index [kg_H2O/kg_fuel]
        - NO : ndarray
            NO emission index [kg_NO/kg_fuel]
        - NO2 : ndarray
            NO2 emission index [kg_NO2/kg_fuel]

    Notes
    -----
    Uses pre-trained surrogate models to estimate emission indices.

    **Extra modules required**

    * numpy

    **Major Assumptions**

    * Operating conditions fall within the training data range
    * Linear interpolation can be employed between training points

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.train_CRN_EI_surrogates
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.build_CRN_EI_surrogates
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera 

    References
    ----------
    [1] Goodwin, D. G., Speth, R. L., Moffat, H. K., & Weber, B. W. (2023). Cantera: An object-oriented software toolkit for chemical kinetics, thermodynamics, and transport processes (Version 3.0.0) [Computer software]. https://www.cantera.org
    """
  
    I          = segment.state.numerics.time.integrate
    surrogates = segment.analyses.emissions.surrogates
    
    CO2_total = 0 * segment.state.ones_row(1)  
    CO_total  = 0 * segment.state.ones_row(1) 
    H2O_total = 0 * segment.state.ones_row(1) 
    NO_total  = 0 * segment.state.ones_row(1) 
    NO2_total = 0 * segment.state.ones_row(1) 

    for network in vehicle.networks:    
        for propulsor in network.propulsors:
            if propulsor.active == True:
                if (type(propulsor) == RCAIDE.Library.Components.Propulsors.Turbofan) or \
                    type(propulsor) == RCAIDE.Library.Components.Propulsors.Turboprop or \
                    type(propulsor) == RCAIDE.Library.Components.Propulsors.Turboshaft or \
                    type(propulsor) == RCAIDE.Library.Components.Propulsors.Turbojet:    
                
                    combustor = propulsor.combustor
                
                    # unpack component conditions
                    propulsor_conditions = segment.state.conditions.energy[propulsor.tag] 
                    combustor_conditions = propulsor_conditions[combustor.tag]  

                    T    = combustor_conditions.inputs.stagnation_temperature
                    P    = combustor_conditions.inputs.stagnation_pressure 
                    mdot = propulsor_conditions.core_mass_flow_rate 
                    FAR  = combustor_conditions.outputs.fuel_to_air_ratio 
                    
                    pts = np.hstack((T,P,mdot,FAR )) 

                    EI_CO2_comb  = np.atleast_2d(surrogates.EI_CO2(pts)).T
                    EI_CO_comb   = np.atleast_2d(surrogates.EI_CO(pts)).T 
                    EI_H2O_comb  = np.atleast_2d(surrogates.EI_H2O(pts)).T 
                    EI_NO_comb   = np.atleast_2d(surrogates.EI_NO(pts)).T 
                    EI_NO2_comb  = np.atleast_2d(surrogates.EI_NO2(pts)).T        
                          
                    CO2_total += np.dot(I,mdot*EI_CO2_comb)
                    CO_total  += np.dot(I,mdot *EI_CO_comb )
                    H2O_total += np.dot(I,mdot*EI_H2O_comb)
                    NO_total  += np.dot(I,mdot *EI_NO_comb ) 
                    NO2_total += np.dot(I,mdot *EI_NO2_comb)

    emissions                 = Data()
    emissions.total           = Data()
    emissions.index           = Data() 
    emissions.total.CO2       = CO2_total * combustor.fuel_data.global_warming_potential_100.CO2 
    emissions.total.H2O       = H2O_total * combustor.fuel_data.global_warming_potential_100.H2O  
    emissions.total.NOx       = (NO_total + NO2_total) * combustor.fuel_data.global_warming_potential_100.NOx 
    emissions.index.CO2       = EI_CO2_comb
    emissions.index.CO        = EI_CO_comb 
    emissions.index.H2O       = EI_H2O_comb
    emissions.index.NO        = EI_NO_comb 
    emissions.index.NO2       = EI_NO2_comb 
 
    segment.state.conditions.emissions =  emissions
    return   

     


# RCAIDE/Library/Missions/Segments/Descent/Constant_CAS_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import Units

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant EAS speed and constant descent rate

    Source:
    N/A

    Inputs:
    segment.equivalent_air_speed                        [meters/second]
    segment.altitude_start                              [meters]
    segment.altitude_end                                [meters]
    segment.descent_rate                                [meters/second]
    segment.state.numerics.dimensionless.control_points [array]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """       
    
    # unpack
    descent_rate = segment.descent_rate
    cas          = segment.calibrated_air_speed   
    alt0         = segment.altitude_start 
    altf         = segment.altitude_end
    beta         = segment.sideslip_angle
    t_nondim     = segment.state.numerics.dimensionless.control_points
    conditions   = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # determine airspeed from calibrated airspeed
    RCAIDE.Library.Mission.Common.Update.atmosphere(segment) # get density for airspeed

    alt_data = segment.analyses.atmosphere.compute_values(alt,segment.temperature_deviation)
    density  = alt_data.density[:,0]  
    pressure = alt_data.pressure[:,0] 

    MSL_data  = segment.analyses.atmosphere.compute_values(0.0,segment.temperature_deviation)
    pressure0 = MSL_data.pressure[0]
    

    if cas is None:
        if not segment.state.initials: raise AttributeError('initial equivalent airspeed not set')
        air_speed =  np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])    
    else:  
        kcas  = cas / Units.knots
        delta = pressure / pressure0 
    
        mach = 2.236*((((1+4.575e-7*kcas**2)**3.5-1)/delta + 1)**0.2857 - 1)**0.5
    
        qc  = pressure * ((1+0.2*mach**2)**3.5 - 1)
        eas = cas * (pressure/pressure0)**0.5*(((qc/pressure+1)**0.286-1)/((qc/pressure0+1)**0.286-1))**0.5
        
        air_speed = eas/np.sqrt(density/MSL_data.density[0])    
    
    # process velocity vector
    v_mag = air_speed
    v_z   = descent_rate # z points down
    v_xy  = np.sqrt( v_mag**2 - v_z**2 )
    v_x   = np.cos(beta)*v_xy
    v_y   = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context


# RCAIDE/Library/Missions/Segments/Ground/Battery_Charge_or_Discharge.py
# 
# 
# Created:  Jul 2023, M. Clarke 
import RCAIDE 
from RCAIDE.Framework.Core import  Units
import  numpy as  np 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------    
def initialize_conditions(segment):  
    """Sets the specified conditions which are given for the segment type.

    Assumptions: 
    During recharging, the charge time associated with the largest capacity battery pack is used 
    
    Source:
    N/A

    Inputs:
    segment.overcharge_contingency              [-]
    battery.                                    [-]

    Outputs: 
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """    
    t_nondim   = segment.state.numerics.dimensionless.control_points

    if isinstance(segment, RCAIDE.Framework.Mission.Segments.Ground.Battery_Recharge):
        for network in segment.analyses.energy.vehicle.networks:
            time =  []
            for bus in  network.busses:
                t=0
                if segment.state.initials.keys():
                    end_of_flight_soc = 1
                    for battery_module in segment.state.conditions.energy[bus.tag].battery_modules:
                        end_of_flight_soc = min(end_of_flight_soc,battery_module.cell.state_of_charge[-1])
                else:
                    end_of_flight_soc =  segment.initial_battery_state_of_charge
                
                t           =  max(((segment.cutoff_SOC-end_of_flight_soc) / bus.charging_c_rate )*Units.hrs  , t) 
                t           += segment.cooling_time
                time.append(t)
            t_initial = segment.state.conditions.frames.inertial.time[0,0]
            t_nondim  = segment.state.numerics.dimensionless.control_points
            time      = np.max(time)
            charging_time      = t_nondim * ( time ) + t_initial 
            segment.state.conditions.frames.inertial.time[:,0] = charging_time[:,0]

    else:

        t_initial = segment.state.conditions.frames.inertial.time[0,0]
        t_nondim  = segment.state.numerics.dimensionless.control_points
        time      = t_nondim * ( segment.time ) + t_initial
        segment.state.conditions.frames.inertial.time[:,0] = time[:,0]


# RCAIDE/Methods/Energy/Propulsors/Electric_Rotor_Propulsor/compute_electric_rotor_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports   
from RCAIDE.Library.Methods.Propulsors.Modulators.Electronic_Speed_Controller.compute_esc_performance  import * 
from RCAIDE.Library.Methods.Propulsors.Converters.DC_Motor.compute_motor_performance                   import *
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.compute_rotor_performance                      import * 


# pacakge imports  
import numpy as np 
from copy import deepcopy

# ----------------------------------------------------------------------------------------------------------------------
# compute_electric_rotor_performance
# ----------------------------------------------------------------------------------------------------------------------  
def compute_electric_rotor_performance(propulsor,state,voltage,center_of_gravity= [[0.0, 0.0,0.0]]):   
    ''' Computes the perfomrance of one propulsor
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure    [-] 
    voltage              - system voltage                         [V]
    bus                  - bus                                    [-] 
    propulsor            - propulsor data structure               [-] 
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W]
    total_current        - current of propulsor group             [A]

    Outputs:  
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W]
    total_current        - current of propulsor group             [A]
    stored_results_flag  - boolean for stored results             [-]     
    stored_propulsor_tag - name of propulsor with stored results  [-]
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                 = state.conditions    
    electric_rotor_conditions  = conditions.energy[propulsor.tag] 
    motor                      = propulsor.motor 
    rotor                      = propulsor.rotor 
    esc                        = propulsor.electronic_speed_controller  
    esc_conditions             = electric_rotor_conditions[esc.tag]
    motor_conditions           = electric_rotor_conditions[motor.tag]
    rotor_conditions           = electric_rotor_conditions[rotor.tag]
    eta                        = conditions.energy[propulsor.tag].throttle
    
    esc_conditions.inputs.voltage   = voltage
    esc_conditions.throttle         = eta 
    compute_voltage_out_from_throttle(esc,esc_conditions,conditions)

    # Assign conditions to the rotor
    motor_conditions.voltage                   = esc_conditions.outputs.voltage 
    compute_RPM_and_torque_from_power_coefficent_and_voltage(motor,motor_conditions,conditions)
    
    # Spin the rotor 
    rotor_conditions.omega           = motor_conditions.omega
    rotor_conditions.throttle        = esc_conditions.throttle 
    compute_rotor_performance(propulsor,state,center_of_gravity)  

    # Run the motor for current
    compute_current_from_RPM_and_voltage(motor,motor_conditions,conditions)
    
    # Detemine esc current 
    esc_conditions.outputs.current = motor_conditions.current
    compute_current_in_from_throttle(esc,esc_conditions,conditions)  
    esc_conditions.current   = esc_conditions.inputs.current  
    esc_conditions.power     = esc_conditions.inputs.power
    
    stored_results_flag     = True
    stored_propulsor_tag    = propulsor.tag 
    
    # compute total forces and moments from propulsor (future work would be to add moments from motors)
    electric_rotor_conditions.thrust      = conditions.energy[propulsor.tag][rotor.tag].thrust 
    electric_rotor_conditions.moment      = conditions.energy[propulsor.tag][rotor.tag].moment 
    
    T  = conditions.energy[propulsor.tag][rotor.tag].thrust 
    M  = conditions.energy[propulsor.tag][rotor.tag].moment 
    P  = conditions.energy[propulsor.tag][esc.tag].power 
    
    return T,M,P, stored_results_flag,stored_propulsor_tag 
                
def reuse_stored_electric_rotor_data(propulsor,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one propulsor for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure    [-] 
    voltage              - system voltage                         [V]
    bus                  - bus                                    [-] 
    propulsors           - propulsor data structure               [-] 
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W]
    total_current        - current of propulsor group             [A]

    Outputs:  
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W]
    total_current        - current of propulsor group             [A] 
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                 = state.conditions 
    motor                      = propulsor.motor 
    rotor                      = propulsor.rotor 
    esc                        = propulsor.electronic_speed_controller  
    motor_0                    = network.propulsors[stored_propulsor_tag].motor 
    rotor_0                    = network.propulsors[stored_propulsor_tag].rotor 
    esc_0                      = network.propulsors[stored_propulsor_tag].electronic_speed_controller
    
    conditions.energy[propulsor.tag][motor.tag]        = deepcopy(conditions.energy[stored_propulsor_tag][motor_0.tag])
    conditions.energy[propulsor.tag][rotor.tag]        = deepcopy(conditions.energy[stored_propulsor_tag][rotor_0.tag])
    conditions.energy[propulsor.tag][esc.tag]          = deepcopy(conditions.energy[stored_propulsor_tag][esc_0.tag])
  
    thrust                  = conditions.energy[propulsor.tag][rotor.tag].thrust 
    power                   = conditions.energy[propulsor.tag][esc.tag].power 
    
    moment_vector           = 0*state.ones_row(3) 
    moment_vector[:,0]      = rotor.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = rotor.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = rotor.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, thrust)
    
    conditions.energy[propulsor.tag][rotor.tag].moment = moment  
    conditions.energy[propulsor.tag].thrust            = thrust   
    conditions.energy[propulsor.tag].moment            = moment  
    
    return thrust,moment,power


# RCAIDE/Library/Components/Propulsors/Constant_Speed_ICE_Propeller.py
# 
#  
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from .                import Propulsor 
from RCAIDE.Library.Methods.Propulsors.Constant_Speed_ICE_Propulsor.append_ice_cs_propeller_conditions  import append_ice_cs_propeller_conditions
from RCAIDE.Library.Methods.Propulsors.Constant_Speed_ICE_Propulsor.compute_cs_ice_performance          import compute_cs_ice_performance, reuse_stored_ice_cs_prop_data
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Constant_Speed_ICE_Propeller
# ---------------------------------------------------------------------------------------------------------------------- 
class Constant_Speed_ICE_Propeller(Propulsor):
    """
    A propulsion system class that combines an internal combustion engine with a constant-speed propeller.
    
    Attributes
    ----------
    tag : str
        Identifier for the propulsion system, defaults to 'ice_constant_speed_propeller'
    
    active_fuel_tanks : list
        List with names of active fuel tanks. Default is None.
        
    engine : None or Engine
        The internal combustion engine component
        
    propeller : None or Propeller
        The constant-speed propeller component
    
    Notes
    -----
    This class models a propulsion system that pairs an internal combustion engine with 
    a constant-speed propeller. The constant-speed propeller maintains a specified RPM 
    by adjusting blade pitch.
    
    **Definitions**

    'Constant-Speed Propeller'
        A propeller that maintains a constant rotational speed by automatically 
        adjusting blade pitch to match power requirements

    'Governor'
        Mechanical or electronic device that controls propeller pitch to maintain 
        desired RPM
    
    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Propulsor
    """ 
    def __defaults__(self):    
        # setting the default values
        self.tag                          = 'ice_constant_speed_propeller'   
        self.active_fuel_tanks            = None
        self.engine                       = None
        self.propeller                    = None  
          

    def append_operating_conditions(self,segment):
        append_ice_cs_propeller_conditions(self,segment)
        return

    def unpack_propulsor_unknowns(self,segment):   
        return 

    def pack_propulsor_residuals(self,segment): 
        return        

    def append_propulsor_unknowns_and_residuals(self,segment): 
        return
        
    def compute_performance(self,state,center_of_gravity = [[0, 0, 0]]):
        thrust,moment,power,stored_results_flag,stored_propulsor_tag =  compute_cs_ice_performance(self,state,center_of_gravity)
        return thrust,moment,power,stored_results_flag,stored_propulsor_tag
    
    def reuse_stored_data(ICE_cs_prop,state,network,stored_propulsor_tag,center_of_gravity = [[0, 0, 0]]):
        thrust,moment,power  = reuse_stored_ice_cs_prop_data(ICE_cs_prop,state,network,stored_propulsor_tag,center_of_gravity)
        return thrust,moment,power           
 



# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/translate_data.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Framework.Core import Units
from .AVL_Objects.Run_Case import Run_Case

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  translate_data
# ---------------------------------------------------------------------------------------------------------------------- 
def translate_conditions_to_cases(avl ,conditions):
    """ Takes RCAIDE Conditions() data structure and translates to a Container of
    avl Run_Case()s.

    Assumptions:
        None
        
    Source:
        Drela, M. and Youngren, H., AVL, http://web.mit.edu/drela/Public/web/avl

    Inputs:
        conditions.aerodynamics.angles.alpha  [radians] 
        conditions.freestream.mach_number        [-]
        conditions.freestream.density            [kilograms per meters**3]
        conditions.freestream.gravity            [meters per second**2]

    Outputs:
        cases                                    [data structur]

    Properties Used:
        N/A
    """    
    # set up aerodynamic Conditions object
    aircraft = avl.vehicle
    cases    = Run_Case.Container()
    for i in range(len(conditions.aerodynamics.angles.alpha)):      
        case                                                  = Run_Case()
        case.tag                                              = avl.settings.filenames.case_template.format(avl.current_status.batch_index,i+1)
        case.mass                                             = conditions.weights.total_mass
        case.conditions.freestream.mach                       = conditions.freestream.mach_number[i, 0]
        case.conditions.freestream.density                    = conditions.freestream.density[i, 0]
        case.conditions.freestream.gravitational_acceleration = conditions.freestream.gravity[i, 0]      
        case.conditions.aerodynamics.angles.alpha             = conditions.aerodynamics.angles.alpha[i, 0]/Units.deg
        case.conditions.aerodynamics.angles.beta              = conditions.aerodynamics.angles.beta[i, 0]/Units.deg 
        if type(conditions.aerodynamics.coefficients.lift.total) == np.ndarray: 
            case.conditions.aerodynamics.coefficients.lift.total= conditions.aerodynamics.coefficients.lift.total[i, 0]        
        else:      
            case.conditions.aerodynamics.coefficients.lift.total = None
        case.conditions.static_stability.coefficients.roll    = conditions.static_stability.coefficients.roll[i, 0] 
        case.conditions.static_stability.coefficients.pitch   = conditions.static_stability.coefficients.pitch[i, 0] 
        
        # determine the number of wings 
        n_wings = 0 
        for wing in aircraft.wings:
            n_wings += 1
            if wing.symmetric == True:
                n_wings += 1                
        case.num_wings                                        = n_wings
        case.n_sw                                             = avl.settings.number_of_spanwise_vortices  
                
        cases.append_case(case)
    
    return cases

def translate_results_to_conditions(cases,res,results):
    """ Takes avl results structure containing the results of each run case stored
        each in its own Data() object. Translates into the Conditions() data structure.

    Assumptions:
        None
        
    Source:
        Drela, M. and Youngren, H., AVL, http://web.mit.edu/drela/Public/web/avl

    Inputs:
        case_res = results 
            
     Outputs:
        cases                        [data_structure]
   
    Properties Used:
        N/A
    """   
   
    num_wings = cases[0].num_wings 
    n_sw      = cases[0].n_sw
    dim       = len(cases)
         
    # aero results 1: total surface forces and coefficeints
    res.aerodynamics.coefficients.lift.total       = np.zeros((dim,1))
    res.aerodynamics.wing_areas                    = np.zeros((dim,num_wings)) 
    res.aerodynamics.wing_CLs                      = np.zeros_like(res.aerodynamics.wing_areas) 
    res.aerodynamics.wing_CDs                      = np.zeros_like(res.aerodynamics.wing_areas) 

    # aero results 2 : sectional forces and coefficients 
    res.aerodynamics.wing_local_spans              = np.zeros((dim,num_wings,n_sw))
    res.aerodynamics.wing_section_chords           = np.zeros_like(res.aerodynamics.wing_local_spans)
    res.aerodynamics.wing_section_cls              = np.zeros_like(res.aerodynamics.wing_local_spans)
    res.aerodynamics.wing_section_induced_angle    = np.zeros_like(res.aerodynamics.wing_local_spans)
    res.aerodynamics.wing_section_cds              = np.zeros_like(res.aerodynamics.wing_local_spans) 
    res.static_stability.control_surfaces_cases   = {}
    
    mach_case = list(results.keys())[0][5:9]   
    for i in range(len(results.keys())):
        aoa_case = '{:04d}'.format(i+1)
        tag = 'case_' + mach_case + '_' + aoa_case
        case_res = results[tag]       
        
        # stability file 
        res.S_ref[i][0]                                                     = case_res.S_ref 
        res.c_ref[i][0]                                                     = case_res.c_ref 
        res.b_ref[i][0]                                                     = case_res.b_ref
        res.X_ref[i][0]                                                     = case_res.X_ref 
        res.Y_ref[i][0]                                                     = case_res.Y_ref 
        res.Z_ref[i][0]                                                     = case_res.Z_ref       
        res.aerodynamics.angles.alpha[i][0]                                 = case_res.aerodynamics.AoA * Units.degree
        res.aerodynamics.angles.beta[i][0]                                  = case_res.aerodynamics.beta * Units.degree
        res.static_stability.coefficients.X[i][0]                           = case_res.aerodynamics.CX 
        res.static_stability.coefficients.Y[i][0]                           = case_res.aerodynamics.CY  
        res.static_stability.coefficients.Z[i][0]                           = case_res.aerodynamics.CZ    
        res.static_stability.coefficients.L[i][0]                           = case_res.aerodynamics.Cltot 
        res.static_stability.coefficients.M[i][0]                           = case_res.aerodynamics.Cmtot 
        res.static_stability.coefficients.N[i][0]                           = case_res.aerodynamics.Cntot      
        res.static_stability.coefficients.roll[i][0]                        = case_res.aerodynamics.roll_moment_coefficient
        res.static_stability.coefficients.pitch[i][0]                       = case_res.aerodynamics.pitch_moment_coefficient
        res.static_stability.coefficients.yaw[i][0]                         = case_res.aerodynamics.yaw_moment_coefficient
        res.aerodynamics.coefficients.lift.total[i][0]                      = case_res.aerodynamics.total_lift_coefficient
        res.aerodynamics.coefficients.drag.induced.inviscid[i][0]           = case_res.aerodynamics.induced_drag_coefficient 
        res.aerodynamics.coefficients.drag.induced.efficiency_factor[i][0]  = case_res.aerodynamics.oswald_efficiency 
        res.aerodynamics.oswald_efficiency[i][0]                            = case_res.aerodynamics.oswald_efficiency
        
        res.static_stability.derivatives.Clift_alpha[i][0]                  = case_res.stability.alpha_derivatives.lift_curve_slope
        res.static_stability.derivatives.CY_alpha[i][0]                     = case_res.stability.alpha_derivatives.side_force_derivative
        res.static_stability.derivatives.CL_alpha[i][0]                     = case_res.stability.alpha_derivatives.roll_moment_derivative
        res.static_stability.derivatives.CM_alpha[i][0]                     = case_res.stability.alpha_derivatives.pitch_moment_derivative
        res.static_stability.derivatives.CN_alpha[i][0]                     = case_res.stability.alpha_derivatives.yaw_moment_derivative
        res.static_stability.derivatives.Clift_beta[i][0]                   = case_res.stability.beta_derivatives.lift_coefficient_derivative
        res.static_stability.derivatives.CY_beta[i][0]                      = case_res.stability.beta_derivatives.side_force_derivative
        res.static_stability.derivatives.CL_beta[i][0]                      = case_res.stability.beta_derivatives.roll_moment_derivative
        res.static_stability.derivatives.CM_beta[i][0]                      = case_res.stability.beta_derivatives.pitch_moment_derivative
        res.static_stability.derivatives.CN_beta[i][0]                      = case_res.stability.beta_derivatives.yaw_moment_derivative        
    
        res.static_stability.derivatives.Clift_p[i][0]                      = case_res.stability.CL_p 
        res.static_stability.derivatives.Clift_q[i][0]                      = case_res.stability.CL_q
        res.static_stability.derivatives.Clift_r[i][0]                      = case_res.stability.CL_r 
        res.static_stability.derivatives.CY_p[i][0]                         = case_res.stability.CY_p 
        res.static_stability.derivatives.CY_q[i][0]                         = case_res.stability.CY_q 
        res.static_stability.derivatives.CY_r[i][0]                         = case_res.stability.CY_r
        res.static_stability.derivatives.CL_p[i][0]                         = case_res.stability.Cl_p 
        res.static_stability.derivatives.CL_q[i][0]                         = case_res.stability.Cl_q 
        res.static_stability.derivatives.CL_r[i][0]                         = case_res.stability.Cl_r 
        res.static_stability.derivatives.CM_p[i][0]                         = case_res.stability.Cm_p 
        res.static_stability.derivatives.CM_q[i][0]                         = case_res.stability.Cm_q 
        res.static_stability.derivatives.CM_r[i][0]                         = case_res.stability.Cm_r 
        res.static_stability.derivatives.CN_p[i][0]                         = case_res.stability.Cn_p 
        res.static_stability.derivatives.CN_q[i][0]                         = case_res.stability.Cn_q 
        res.static_stability.derivatives.CN_r[i][0]                         = case_res.stability.Cn_r
        res.static_stability.derivatives.CX_u[i][0]                         = case_res.stability.CX_u
        res.static_stability.derivatives.CX_v[i][0]                         = case_res.stability.CX_v
        res.static_stability.derivatives.CX_w[i][0]                         = case_res.stability.CX_w
        res.static_stability.derivatives.CY_u[i][0]                         = case_res.stability.CY_u
        res.static_stability.derivatives.CY_v[i][0]                         = case_res.stability.CY_v
        res.static_stability.derivatives.CY_w[i][0]                         = case_res.stability.CY_w
        res.static_stability.derivatives.CZ_u[i][0]                         = case_res.stability.CZ_u
        res.static_stability.derivatives.CZ_v[i][0]                         = case_res.stability.CZ_v
        res.static_stability.derivatives.CZ_w[i][0]                         = case_res.stability.CZ_w
        res.static_stability.derivatives.CL_u[i][0]                         = case_res.stability.Cl_u
        res.static_stability.derivatives.CL_v[i][0]                         = case_res.stability.Cl_v
        res.static_stability.derivatives.CL_w[i][0]                         = case_res.stability.Cl_w
        res.static_stability.derivatives.CM_u[i][0]                         = case_res.stability.Cm_u
        res.static_stability.derivatives.CM_v[i][0]                         = case_res.stability.Cm_v
        res.static_stability.derivatives.CM_w[i][0]                         = case_res.stability.Cm_w
        res.static_stability.derivatives.CN_u[i][0]                         = case_res.stability.Cn_u
        res.static_stability.derivatives.CN_v[i][0]                         = case_res.stability.Cn_v
        res.static_stability.derivatives.CN_w[i][0]                         = case_res.stability.Cn_w
            
        res.static_stability.derivatives.CX_p[i][0]                         = case_res.stability.CX_p
        res.static_stability.derivatives.CX_q[i][0]                         = case_res.stability.CX_q
        res.static_stability.derivatives.CX_r[i][0]                         = case_res.stability.CX_r
        res.static_stability.derivatives.CY_p[i][0]                         = case_res.stability.CY_p
        res.static_stability.derivatives.CY_q[i][0]                         = case_res.stability.CY_q
        res.static_stability.derivatives.CY_r[i][0]                         = case_res.stability.CY_r
        res.static_stability.derivatives.CZ_p[i][0]                         = case_res.stability.CZ_p
        res.static_stability.derivatives.CZ_q[i][0]                         = case_res.stability.CZ_q
        res.static_stability.derivatives.CZ_r[i][0]                         = case_res.stability.CZ_r   
        res.static_stability.neutral_point[i][0]                            = case_res.stability.neutral_point
        res.static_stability.spiral_criteria[i][0]                          = case_res.stability.spiral_criteria
        
        # aero surface forces file 
        res.aerodynamics.wing_areas[i][:]                   = case_res.aerodynamics.wing_areas   
        res.aerodynamics.wing_CLs[i][:]                     = case_res.aerodynamics.wing_CLs    
        res.aerodynamics.wing_CDs[i][:]                     = case_res.aerodynamics.wing_CDs    
        
        # aero sectional forces file
        res.aerodynamics.wing_local_spans[i][:]             = case_res.aerodynamics.wing_local_spans
        res.aerodynamics.wing_section_chords[i][:]          = case_res.aerodynamics.wing_section_chords  
        res.aerodynamics.wing_section_cls[i][:]             = case_res.aerodynamics.wing_section_cls    
        res.aerodynamics.wing_section_induced_angle[i][:]   = case_res.aerodynamics.wing_section_aoa_i
        res.aerodynamics.wing_section_cds[i][:]             = case_res.aerodynamics.wing_section_cds   
        
        res.static_stability.control_surfaces_cases[tag]    = case_res.stability.control_surfaces
        
    if len(res.static_stability.coefficients.X) > 1:
        res.static_stability.derivatives.CX_alpha[:, 0] =  np.gradient( res.static_stability.coefficients.X[:, 0],res.aerodynamics.angles.alpha[:, 0] )
        res.static_stability.derivatives.CZ_alpha[:, 0] =  np.gradient( res.static_stability.coefficients.Z[:, 0],res.aerodynamics.angles.alpha[:, 0] )
    return  



# RCAIDE/Methods/Aerodynamics/Common/Drag/parasite_drag_wing.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
from RCAIDE.Framework.Core import Data  
from RCAIDE.Library.Methods.Aerodynamics.Common.Drag.compressible_mixed_flat_plate import compressible_mixed_flat_plate
from RCAIDE.Library.Methods.Utilities         import Cubic_Spline_Blender  
 
# package imports
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------  
#   Parasite Drag Wing 
# ----------------------------------------------------------------------------------------------------------------------   
def parasite_drag_wing(state,settings,geometry):
    """Computes the parasite drag due to wings

    Assumptions:
    Basic fit

    Source:
    http://aerodesign.stanford.edu/aircraftdesign/aircraftdesign.html (Stanford AA241 A/B Course Notes)

    Args:
    settings.wing_parasite_drag_form_factor      [Unitless]
    state.conditions.freestream.
      mach_number                                [Unitless]
      temperature                                [K]
      reynolds_number                            [Unitless]
    geometry.
      areas.reference                            [m^2]
      chords.mean_aerodynamic                    [m]
      thickness_to_chord                         [Unitless]
      sweeps.quarter_chord                       [radians]
      aspect_ratio                               [Unitless]
      spans.projected                            [m]
      areas.affected                             [m^2]
      areas.wetted                               [m^2]
      transition_x_upper                         [Unitless]
      transition_x_lower                         [Unitless]
      
      
    Returns:
    wing_parasite_drag                           [Unitless] 
    """
    
    # unpack inputs
    C                             = settings.wing_parasite_drag_form_factor 
    recalculate_total_wetted_area = settings.recalculate_total_wetted_area
    freestream                    = state.conditions.freestream
    
    # conditions
    Mc  = freestream.mach_number
    Tc  = freestream.temperature    
    re  = freestream.reynolds_number     
    
    wing = geometry
    wing_parasite_drag = 0.0
    
    # Unpack wing
    exposed_root_chord_offset = wing.exposed_root_chord_offset
    t_c_w                     = wing.thickness_to_chord
    Sref                      = wing.areas.reference
    num_segments              = len(wing.segments.keys())     
    
    # if wing has segments, compute and sum parasite drag of each segment

    xtu       = wing.transition_x_upper
    xtl       = wing.transition_x_lower     
    
    if num_segments>0:        
        total_segment_parasite_drag  = 0 
        total_segment_k_w            = 0 
        total_segment_cf_w_u         = 0
        total_segment_cf_w_l         = 0 
        total_segment_k_comp_u       = 0
        total_segment_k_comp_l       = 0
        total_k_reyn_u               = 0          
        total_k_reyn_l               = 0
        
        for i,segment in enumerate (wing.segments): 
            if i == num_segments-1:
                continue  
            mac_seg       = segment.chords.mean_aerodynamic
            Sref_seg      = segment.areas.reference 
            Swet_seg      = segment.areas.wetted
            sweep_seg     = segment.sweeps.quarter_chord
    
            # compute parasite drag coef., form factor, skin friction coef., compressibility factor and reynolds number for segments
            segment_parasite_drag , segment_k_w, segment_cf_w_u, segment_cf_w_l, segment_k_comp_u, segment_k_comp_l, k_reyn_u ,k_reyn_l = compute_parasite_drag(re,mac_seg,Mc,Tc,xtu,xtl,sweep_seg,t_c_w,Sref_seg,Swet_seg,C)    
            
            total_segment_parasite_drag  += segment_parasite_drag*Sref_seg   
            total_segment_k_w            += segment_k_w*Sref_seg 
            total_segment_cf_w_u         += segment_cf_w_u*Sref_seg 
            total_segment_cf_w_l         += segment_cf_w_l*Sref_seg 
            total_segment_k_comp_u       += segment_k_comp_u*Sref_seg 
            total_segment_k_comp_l       += segment_k_comp_l*Sref_seg 
            total_k_reyn_u               += k_reyn_u*Sref_seg                 
            total_k_reyn_l               += k_reyn_l*Sref_seg  
                
        wing_parasite_drag = total_segment_parasite_drag  / Sref
        k_w                = total_segment_k_w / Sref
        cf_w_u             = total_segment_cf_w_u  / Sref
        cf_w_l             = total_segment_cf_w_l / Sref
        k_comp_u           = total_segment_k_comp_u  / Sref
        k_comp_l           = total_segment_k_comp_l  / Sref
        k_reyn_u           = total_k_reyn_u  / Sref
        k_reyn_l           = total_k_reyn_l  / Sref

    # if wing has no segments      
    else:              
        # wing
        mac_w        = wing.chords.mean_aerodynamic
        sweep_w      = wing.sweeps.quarter_chord 
        span_w       = wing.spans.projected
        Sref         = wing.areas.reference
        
        chord_root = wing.chords.root
        chord_tip  = wing.chords.tip
        wing_root     = chord_root + exposed_root_chord_offset*((chord_tip - chord_root)/span_w)
        
        if recalculate_total_wetted_area or wing.areas.wetted==0.:  
            
            # calculate exposed area
            if wing.symmetric:
                S_exposed_w = wing.areas.reference - (chord_root + wing_root)*exposed_root_chord_offset         
            else: 
                S_exposed_w = wing.areas.reference - 0.5*(chord_root + wing_root)*exposed_root_chord_offset
                
            if t_c_w < 0.05:
                Swet = 2.003* S_exposed_w
            else:
                Swet = (1.977 + 0.52*t_c_w) * S_exposed_w   
                
            wing.areas.wetted = Swet 
        else:
            Swet              = wing.areas.wetted                         

        # compute parasite drag coef., form factor, skin friction coef., compressibility factor and reynolds number for wing
        wing_parasite_drag , k_w, cf_w_u, cf_w_l, k_comp_u, k_comp_l, k_reyn_u, k_reyn_l = compute_parasite_drag(re,mac_w,Mc,Tc,xtu,xtl,sweep_w,t_c_w,Sref,Swet,C)             

    # dump data to conditions
    wing_result = Data(
        wetted_area               = wing.areas.wetted,
        reference_area            = Sref   , 
        total                     = wing_parasite_drag ,
        skin_friction             = (cf_w_u+cf_w_l)/2.   ,
        compressibility_factor    = (k_comp_u+k_comp_l)/2 ,
        reynolds_factor           = (k_reyn_u+k_reyn_l)/2 , 
        form_factor               = k_w    ,
    )
    
    state.conditions.aerodynamics.coefficients.drag.parasite[wing.tag] = wing_result

    return  
 
def compute_parasite_drag(re,mac_w,Mc,Tc,xtu,xtl,sweep_w,t_c_w,Sref,Swet,C):
    """Computes the parasite drag due to wings

    Assumptions:
    Basic fit

    Source:
    adg.stanford.edu (Stanford AA241 A/B Course Notes)

    Args:
    re (Reynolds Number)    [Unitless]
    mac_w (Wing MAC)        [m]
    Mc (Mach Number)        [Unitless]
    Tc (Temperature)        [K]
    xtu (Upper Transition)  [Unitless] (percent of chord)
    xtl (Lower Transition)  [Unitless] (percent of chord)
    sweep_w (Wing Sweep)    [rad]
    t_c_w (Wing t/c)        [Unitless]
    Sref (Wing Ref Area)    [m^2]
    Swet (Wing Wetted Area) [m^2]
    C (Form Factor)         [Unitless]
      
    Returns:
    (u is upper, l is lower)
    wing_parasite_drag                 [Unitless]
    k_w    (Form Factor)               [Unitless]
    cf_w_u (Skin Friction Coefficient) [Unitless]
    cf_w_l                             [Unitless]
    k_comp_u (Compressibility Factor)  [Unitless]
    k_comp_l                           [Unitless]
    k_reyn_u (Reynolds Factor)         [Unitless]
    k_reyn_l                           [Unitless] 
    """    
   
    # reynolds number
    Re_w = re*mac_w  
    
    # skin friction  coefficient, upper
    cf_w_u, k_comp_u, k_reyn_u = compressible_mixed_flat_plate(Re_w,Mc,Tc,xtu)
    
    # skin friction  coefficient, lower
    cf_w_l, k_comp_l, k_reyn_l = compressible_mixed_flat_plate(Re_w,Mc,Tc,xtl) 
    
    # correction for airfoils
    cos_sweep = np.cos(sweep_w)
    cos2      = cos_sweep*cos_sweep
    
    ind = Mc <= 1.
    
    k_w = np.ones_like(Mc)
    k_w[ind] = 1. + ( 2.* C * (t_c_w * cos2) ) / ( np.sqrt(1.- Mc[ind]*Mc[ind] * cos2) )  \
            + ( C*C * cos2 * t_c_w*t_c_w * (1. + 5.*(cos2)) ) \
            / (2.*(1.-(Mc[ind]*cos_sweep)**2.))             
    
    spline = Cubic_Spline_Blender(.95,1.0)
    h00 = lambda M:spline.compute(M)
    
    k_w = k_w*(h00(Mc)) + 1*(1-h00(Mc))   

    # find the final result
    wing_parasite_drag = k_w * cf_w_u * Swet / Sref /2. + k_w * cf_w_l * Swet / Sref /2.


    return wing_parasite_drag , k_w, cf_w_u, cf_w_l, k_comp_u, k_comp_l, k_reyn_u, k_reyn_l


# RCAIDE/Library/Methods/Propulsors/Converters/Ducted_Fan/__init__.py
# (c) Copyright 2023 Aerospace Research Community LLC

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .purge_files                        import purge_files
from .read_results                       import read_results
from .run_dfdc_analysis                  import run_dfdc_analysis
from .translate_conditions_to_dfdc_cases import translate_conditions_to_dfdc_cases
from .compute_ducted_fan_performance     import compute_ducted_fan_performance
from .append_ducted_fan_conditions       import append_ducted_fan_conditions
from .write_geometry                     import write_geometry
from .write_input_deck                   import write_input_deck
from .design_ducted_fan                  import design_ducted_fan


# RCAIDE/Library/Methods/Propulsors/Converters/Compression_Nozzle/compute_compression_nozzle_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------     

# package imports
import numpy as np  
from warnings import warn

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_compression_nozzle_performance
# ----------------------------------------------------------------------------------------------------------------------    
def compute_compression_nozzle_performance(compression_nozzle,nozzle_conditions,conditions):
    """  Computes the performance of a compression nozzle bases on its polytropic efficiency.
         The following properties are computed: 
        compression_nozzle.outputs.
          stagnation_temperature             (numpy.ndarray): exit stagnation temperature [K]
          stagnation_pressure                (numpy.ndarray): exit stagnation pressure    [Pa]
          stagnation_enthalpy                (numpy.ndarray): exit stagnation enthalpy    [J/kg]
          mach_number                        (numpy.ndarray): exit Mach number            [unitless]
          static_temperature                 (numpy.ndarray): exit static temperature     [K]
          static_enthalpy                    (numpy.ndarray): exit static enthalpy        [J/kg]
          velocity                           (numpy.ndarray): exit nozzle velocity        [m/s]
          

    Assumptions:  
        1. Pressure ratio and polytropic efficiency do not change with varying conditions.
        2. Adiabatic
        3. Subsonic or choked output.

    Source:
        https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Args:
        conditions.freestream.
          isentropic_expansion_factor         (numpy.ndarray): isentropic_expansion_factor        [unitless]
          specific_heat_at_constant_pressure  (numpy.ndarray): specific_heat_at_constant_pressure [J/(kg K)]
          pressure                            (numpy.ndarray): pressure                           [Pa]
          gas_specific_constant               (numpy.ndarray): gas_specific_constant              [J/(kg K)]
        compression_nozzle.
          inputs.stagnation_temperature       (numpy.ndarra): entering stagnation_temperature     [K]
          inputs.stagnation_pressure          (numpy.ndarra): entering stagnation_pressure        [Pa] 
          pressure_ratio                             (float): pressure_ratio                      [unitless]
          polytropic_efficiency                      (float): polytropic_efficiency               [unitless]
          pressure_recovery                          (float): pressure_recovery                   [unitless]

    Returns:
    """

    # Unpack conditions
    #gamma   = conditions.freestream.isentropic_expansion_factor
    #Cp      = conditions.freestream.specific_heat_at_constant_pressure
    P0      = conditions.freestream.pressure
    M0      = conditions.freestream.mach_number 

    # Unpack inpust
    Tt_in                   = nozzle_conditions.inputs.stagnation_temperature
    Pt_in                   = nozzle_conditions.inputs.stagnation_pressure
    PR                      = compression_nozzle.pressure_ratio
    eta_p_old               = compression_nozzle.polytropic_efficiency
    eta_rec                 = compression_nozzle.pressure_recovery
    compressibility_effects = compression_nozzle.compressibility_effects
    
    # Unpack ram inputs
    working_fluid           = compression_nozzle.working_fluid
 
    # Compute the working fluid properties  
    T0     = nozzle_conditions.inputs.static_temperature
    P0     = nozzle_conditions.inputs.static_pressure   
    M0     = nozzle_conditions.inputs.mach_number 
    gamma  = working_fluid.compute_gamma(T0,P0) 
    Cp     = working_fluid.compute_cp(T0,P0)  
    a      = working_fluid.compute_speed_of_sound(T0,P0)    
 
    # Compute output stagnation quantities
    Pt_out  = Pt_in*PR*eta_rec
    Tt_out  = Tt_in*(PR*eta_rec)**((gamma-1)/(gamma*eta_p_old))
    ht_out  = Tt_out*Cp 

    if compressibility_effects: 
        # initilize arrays
        Pt_out   = np.ones_like(Tt_in)
        M_out     = np.ones_like(Tt_in)
        T_out    = np.ones_like(Tt_in)
        P_out    = np.ones_like(Tt_in)

        if M_out <= 1.0: # use isentropic relations
            i_low          = M0 <= 1.0
            Pt_out[i_low]  = Pt_in[i_low]*PR
            M_out[i_low]    = np.sqrt( (((Pt_out[i_low]/P0[i_low])**((gamma[i_low]-1.)/gamma[i_low]))-1.) *2./(gamma[i_low]-1.) ) 
            T_out[i_low]   = Tt_out[i_low]/(1.+(gamma[i_low]-1.)/2.*M_out[i_low]*M_out[i_low])
            P_out[i_low]   = Pt_out[i_low]/((1.+(gamma[i_low]-1.)/2.*M_out[i_low]*M_out[i_low])**(gamma[i_low]/(gamma[i_low]-1.)))

        else: # use normal shock
            i_high         = M0 > 1.0
            M_out[i_high]   = np.sqrt((1.+(gamma[i_high]-1.)/2.*M0[i_high]**2.)/(gamma[i_high]*M0[i_high]**2-(gamma[i_high]-1.)/2.))
            T_out[i_high]  = Tt_out[i_high]/(1.+(gamma[i_high]-1.)/2*M_out[i_high]*M_out[i_high])
            Pt_out[i_high] = PR*Pt_in[i_high]*((((gamma[i_high]+1.)*(M0[i_high]**2.))/((gamma[i_high]-1.)*\
                            M0[i_high]**2.+2.))**(gamma[i_high]/(gamma[i_high]-1.)))*((gamma[i_high]+1.)/(2.*gamma[i_high]*\
                            M0[i_high]**2.-(gamma[i_high]-1.)))**(1./(gamma[i_high]-1.))
            P_out[i_high]  = Pt_out[i_high]/(1.+(gamma[i_high]-1.)/2.*M_out[i_high]**2.)**(gamma[i_high]/(gamma[i_high]-1.))
    else:
        Pt_out  = Pt_in*PR*eta_rec 
        if np.any(Pt_out<P0): # in case pressures go too low
            warn('Pt_out goes too low',RuntimeWarning)
            Pt_out[Pt_out<P0] = P0[Pt_out<P0] 
        M_out   = np.sqrt( (((Pt_out/P0)**((gamma-1.)/gamma))-1.) *2./(gamma-1.) )
        T_out  = Tt_out/(1.+(gamma-1.)/2.*M_out*M_out)
        P_out  = Pt_out/(1.+(gamma-1.)/2.*M_out*M_out)**(gamma/(gamma-1.))
        
    # Compute exit ethalpy and velocity  
    h_out   = Cp*T_out
    u_out   = np.sqrt(2.*(ht_out-h_out))

    # Pack computed quantities into outputs
    nozzle_conditions.outputs.mach_number             = M_out
    nozzle_conditions.outputs.velocity                = u_out
    nozzle_conditions.outputs.static_enthalpy         = h_out
    nozzle_conditions.outputs.static_temperature      = T_out
    nozzle_conditions.outputs.static_pressure         = P_out
    nozzle_conditions.outputs.stagnation_enthalpy     = ht_out
    nozzle_conditions.outputs.stagnation_temperature  = Tt_out
    nozzle_conditions.outputs.stagnation_pressure     = Pt_out
    
    return 


# RCAIDE/Library/Methods/Propulsors/Converters/Motor/append_motor_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_motor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_motor_conditions(motor,segment,propulsor_conditions): 
    ones_row    = segment.state.ones_row 
    propulsor_conditions[motor.tag]            = Conditions()
    propulsor_conditions[motor.tag].inputs     = Conditions()
    propulsor_conditions[motor.tag].outputs    = Conditions()
    propulsor_conditions[motor.tag].torque     = 0. * ones_row(1) 
    propulsor_conditions[motor.tag].efficiency = 0. * ones_row(1) 
    
    return 



# RCAIDE/Methods/Mission/Segments/Cruise/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from . import Constant_Mach_Constant_Altitude
from . import Constant_Speed_Constant_Altitude
from . import Constant_Mach_Constant_Altitude_Loiter
from . import Constant_Throttle_Constant_Altitude
from . import Variable_Cruise_Distance
from . import Constant_Dynamic_Pressure_Constant_Altitude_Loiter
from . import Constant_Acceleration_Constant_Altitude
from . import Constant_Pitch_Rate_Constant_Altitude
from . import Constant_Dynamic_Pressure_Constant_Altitude
from . import Constant_Speed_Constant_Altitude_Loiter
from . import Curved_Constant_Radius_Constant_Speed_Constant_Altitude



# RCAIDE/Methods/Noise/Boom/lift_equivalent_area.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
from RCAIDE.Library.Methods.Aerodynamics.Vortex_Lattice_Method import VLM

# Python package imports   
import numpy as np  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Equivalent Area from lift for Sonic Boom
# ----------------------------------------------------------------------------------------------------------------------         
def lift_equivalent_area(config,analyses,conditions):
    """ This method calculates the lift equivalent area for a vehicle for sonic boom analysis.
    
        Assumptions:
        X_locs is the location where lift values are taken on the x-axis
        AE_x is the lift equivalent area
        CP is the coefficient of pressure at the flight conditions
        
        
        Source:
        N/A
        
        Inputs:
        config                                [vehicle]
        conditions
              .freestream.dynamic_pressure    [pascals]
              .freestream.mach_number         [-]
              .aerodynamics.angles.alpha   [radians]
        analyses.aerodynamics.settings        [needed for discretization]
        config.total_length                   [m]

        Outputs:
        X_locs                                [m]
        AE_x                                  [m^2]
        CP                                    [pascals]
        
        Properties Used:
        N/A
    """       
    
    S            = config.reference_area
    q            = conditions.freestream.dynamic_pressure
    mach         = conditions.freestream.mach_number
    aoa          = conditions.aerodynamics.angles.alpha
    settings     = analyses.aerodynamics.process.compute.lift.inviscid_wings.settings
    length       = config.total_length
    
    results = VLM(conditions, settings, config)
    CP = results.CP 
    VD = results.VD
    
    areas      = VD.panel_areas
    normal_vec = VD.normals
    XC         = VD.XC
    ZC         = VD.ZC
    z_comp     = normal_vec[:,2]

    lift_force_per_panel = CP[0,:]*q*z_comp*areas*np.cos(aoa[0])
    
    # Mach angle
    mach_angle = np.arcsin(1/mach[0])
    
    # Shift all points onto the X axis
    X_shift = XC + ZC*np.tan(mach_angle)
    
    # Order the values
    sort_order = np.argsort(X_shift)
    X  = np.take(X_shift,sort_order)
    Y  = np.take(lift_force_per_panel, sort_order)

    u, inv = np.unique(X, return_inverse=True)
    sums   = np.zeros(len(u), dtype=Y.dtype) 
    np.add.at(sums, inv, Y) 
    
    lift_forces = sums
    X_locations = u
    
    summed_lift_forces = np.cumsum(lift_forces)
    
    beta = np.sqrt(conditions.freestream.mach_number[0][0]**2 -1)
    
    Ae_lift_at_each_x = (beta/(2*q[0]))*summed_lift_forces
    
    X_max  = length*1.25
    
    X_locs = np.concatenate([[0],X_locations,[X_max]])
    AE_x   = np.concatenate([[0],Ae_lift_at_each_x,[Ae_lift_at_each_x[-1]]])
    
    return X_locs, AE_x, CP


# RCAIDE/Library/Missions/Common/Initialize/time.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Time
# ---------------------------------------------------------------------------------------------------------------------- 
def time(segment):
    """ Sets the initial time of the vehicle at the start of the segment
    
        Assumptions:
        Only used if there is an initial condition
        
        Inputs:
            state.initials.conditions:
                frames.inertial.time     [seconds]
                frames.planet.start_time [seconds]
            state.conditions:           
                frames.inertial.time     [seconds]
            segment.start_time           [seconds]
            
        Outputs:
            state.conditions:           
                frames.inertial.time     [seconds]
                frames.planet.start_time [seconds]

        Properties Used:
        N/A
                                
    """        
    
    if segment.state.initials:
        t_initial = segment.state.initials.conditions.frames.inertial.time
        t_current = segment.state.conditions.frames.inertial.time 
        segment.state.conditions.frames.inertial.time[:,:] = t_current + (t_initial[-1,0] - t_current[0,0])
        
    else:
        t_initial = segment.state.conditions.frames.inertial.time[0,0]
        
    if segment.state.initials:
        segment.state.conditions.frames.planet.start_time = segment.state.initials.conditions.frames.planet.start_time
        
    elif 'start_time' in segment:
        segment.state.conditions.frames.planet.start_time = segment.start_time
    
    return 



# RCAIDE/Core/Container.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------        

from .               import Data
from warnings        import warn
import random

import string
chars = string.punctuation + string.whitespace
t_table = str.maketrans( chars          + string.ascii_uppercase , 
                            '_'*len(chars) + string.ascii_lowercase )

# ----------------------------------------------------------------------------------------------------------------------
#  Container
# ----------------------------------------------------------------------------------------------------------------------   

class Container(Data):
    # """ A dict-type container with attribute, item and index style access
    #     intended to hold a attribute-accessible list of Data(). This is unordered.
        
    #     Assumptions:
    #     N/A
        
    #     Source:
    #     N/A
        
    # """
            
        
    def __defaults__(self):
        # """ Defaults function
    
        #     Assumptions:
        #     None
        
        #     Source:
        #     N/A
        
        #     Inputs:
        #     N/A
        
        #     Outputs:
        #     N/A
            
        #     Properties Used:
        #     N/A
        # """          
        pass
    
    def __init__(self,*args,**kwarg):
        # """ Initialization that builds the container
        
        #     Assumptions:
        #     None
        
        #     Source:
        #     N/A
        
        #     Inputs:
        #     self
        
        #     Outputs:
        #     N/A
            
        #     Properties Used:
        #     N/A
        # """          
        super(Container,self).__init__(*args,**kwarg)
        self.__defaults__()
    
    def append(self,val):
        # """ Appends the value to the containers
        #     This overrides the Data class append by allowing for duplicate named components
        #     The following components will get new names.
        
        #     Assumptions:
        #     None
        
        #     Source:
        #     N/A
        
        #     Inputs:
        #     self
        
        #     Outputs:
        #     N/A
            
        #     Properties Used:
        #     N/A
        # """           
        
        # See if the item tag exists, if it does modify the name
        keys = self.keys()
        
        tag = str.lower(val.tag.translate(t_table))
        if tag in keys:
            string_of_keys = "".join(self.keys())
            n_comps = string_of_keys.count(val.tag)
            val.tag = tag + str(n_comps+1)
            
            # Check again, because theres an outside chance that its duplicate again. Then assign a random
            if val.tag in keys:
                val.tag = tag + str(n_comps+random.randint(0,1000))
        
        Data.append(self,val)
        
    def extend(self,vals):
        # """ Append things regressively depending on what is inside.
    
        #     Assumptions:
        #     None
        
        #     Source:
        #     N/A
        
        #     Inputs:
        #     self
        
        #     Outputs:
        #     N/A
            
        #     Properties Used:
        #     N/A
        # """         
        if isinstance(vals,(list,tuple)):
            for v in val: self.append(v)
        elif isinstance(vals,dict):
            self.update(vals)
        else:
            raise Exception('unrecognized data type')
        
    def get_children(self):
        # """ Returns the components that can go inside
        
        # Assumptions:
        # None
    
        # Source:
        # N/A
    
        # Inputs:
        # None
    
        # Outputs:
        # None
    
        # Properties Used:
        # N/A
        # """        
        
        return []    


# RCAIDE/Library/Missions/Segments/Climb/Constant_Dynamic_Pressure_Constant_Angle.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE 
import RCAIDE 

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions_unpack_unknowns(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constrant dynamic pressure and constant rate of climb

    Source:
    N/A

    Inputs:
    segment.climb_angle                                 [radians]
    segment.dynamic_pressure                            [pascals]
    segment.altitude_start                              [meters]
    segment.altitude_end                                [meters]
    segment.state.numerics.dimensionless.control_points [unitless]
    conditions.freestream.density                       [kilograms/meter^3]  

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.energy.throttle              [unitless]
    conditions.frames.body.inertial_rotations   [radians]

    Properties Used:
    N/A
    """           
    
    # unpack
    climb_angle = segment.climb_angle
    q           = segment.dynamic_pressure
    alt0        = segment.altitude_start  
    conditions  = segment.state.conditions
    beta        = segment.sideslip_angle
    rho         = conditions.freestream.density[:,0]  
    
    # unpack unknowns  
    alts     = conditions.frames.inertial.position_vector[:,2]

    # Update freestream to get density
    RCAIDE.Library.Mission.Common.Update.atmosphere(segment)
    rho = conditions.freestream.density[:,0]   

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # pack conditions    
    conditions.freestream.altitude[:,0] =  -alts  # positive altitude in this context    
    

    # check for initial velocity
    if q is None: 
        if not segment.state.initials: raise AttributeError('dynamic pressure not set')
        v_mag = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
    else: 
        # Update freestream to get density
        RCAIDE.Library.Mission.Common.Update.atmosphere(segment)
        rho = conditions.freestream.density[:,0]       
    
        # process velocity vector
        v_mag = np.sqrt(2*q/rho)
        
    v_x   = np.cos(beta)*v_mag * np.cos(climb_angle)
    v_y   = np.sin(beta)*v_mag * np.cos(climb_angle)
    v_z   = -v_mag * np.sin(climb_angle)
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z   
    
def residual_altitude(segment):
    """Computes the altitude residual

    Assumptions:
    No higher order terms.

    Source:
    N/A

    Inputs:
    segment.state.conditions.frames.inertial.total_force_vector   [Newtons]
    segment.state.conditions.frames.inertial.acceleration_vector  [meter/second^2]
    segment.state.conditions.weights.total_mass                   [kilogram]
    segment.state.conditions.freestream.altitude                  [meter]

    Outputs:
    segment.state.residuals.altitude                              [meters] 

    Properties Used:
    N/A
    """     
    
    # Unpack results 
    alt_in  = segment.state.unknowns.altitude[:,0] 
    alt_out = segment.state.conditions.freestream.altitude[:,0]  
    segment.state.residuals.altitude[:,0] = (alt_in - alt_out)/alt_out[-1]

    return


# ----------------------------------------------------------------------------------------------------------------------  
# Update Differentials
# ----------------------------------------------------------------------------------------------------------------------     
def update_differentials(segment):
    """ On each iteration creates the differentials and integration functions from knowns about the problem. 
      Sets the time at each point. Must return in dimensional time, with t[0] = 0.
      This is different from the common method as it also includes the scaling of operators.

        Assumptions:
        Works with a segment discretized in vertical position, altitude

        Inputs:
        state.numerics.dimensionless.control_points      [Unitless]
        state.numerics.dimensionless.differentiate       [Unitless]
        state.numerics.dimensionless.integrate           [Unitless]
        state.conditions.frames.inertial.position_vector [meter]
        state.conditions.frames.inertial.velocity_vector [meter/second]
        

        Outputs:
        state.conditions.frames.inertial.time            [second]

    """    

    # unpack
    numerics   = segment.state.numerics
    conditions = segment.state.conditions
    x          = numerics.dimensionless.control_points
    D          = numerics.dimensionless.differentiate
    I          = numerics.dimensionless.integrate 
    r          = segment.state.conditions.frames.inertial.position_vector
    v          = segment.state.conditions.frames.inertial.velocity_vector
    alt0       = segment.altitude_start
    altf       = segment.altitude_end    

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
        
    dz = altf - alt0
    vz = -v[:,2,None] # maintain column array

    # get overall time step
    dt = (dz/np.dot(I,vz))[-1]

    # rescale operators
    x = x * dt
    D = D / dt
    I = I * dt
    
    # Calculate the altitudes
    alt = np.dot(I,vz) + alt0
    
    # pack
    t_initial                                       = segment.state.conditions.frames.inertial.time[0,0]
    numerics.time.control_points                    = x
    numerics.time.differentiate                     = D
    numerics.time.integrate                         = I
    conditions.frames.inertial.time[1:,0]            = t_initial + x[1:,0]
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0]  
    conditions.freestream.altitude[:,0]             =  alt[:,0]  

    return


# RCAIDE/Library/Missions/Segments/Climb/Constant_Mach_Constant_Angle.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ----------------------------------------------------------------------------------------------------------------------  
# import RCAIDE 
from RCAIDE.Library.Mission.Common.Update.atmosphere import atmosphere

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.
    
    Assumptions:
    Constant Mach number, with a constant angle of climb

    Source:
    N/A

    Inputs:
    segment.climb_angle                                 [radians]
    segment.mach_number                                 [Unitless]
    segment.altitude_start                              [meters]
    segment.altitude_end                                [meters]
    segment.state.numerics.dimensionless.control_points [Unitless]
    conditions.freestream.density                       [kilograms/meter^3]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]

    Properties Used:
    N/A
    """       
    # unpack User Inputs
    climb_angle = segment.climb_angle
    mach_number = segment.mach_number
    alt0        = segment.altitude_start 
    beta        = segment.sideslip_angle 
    conditions  = segment.state.conditions  
        
    # unpack unknowns  
    alts     = conditions.frames.inertial.position_vector[:,2]
    
    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # pack conditions   
    conditions.freestream.altitude[:,0]   = -alts 

    # check for initial velocity
    if mach_number is None: 
        if not segment.state.initials: raise AttributeError('mach not set')
        v_mag  = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])*segment.state.ones_row(1)   
    else: 
        # Update freestream to get speed of sound
        atmosphere(segment)
        a = conditions.freestream.speed_of_sound    
        
        # process velocity vector
        v_mag = mach_number * a
    v_xy  = v_mag * np.cos(climb_angle)
    v_z   = -v_mag * np.sin(climb_angle)
    v_x   = np.cos(beta)*v_xy
    v_y   = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0]              = v_x[:,0]
    conditions.frames.inertial.velocity_vector[:,1]              = v_y[:,0]
    conditions.frames.inertial.velocity_vector[:,2]              = v_z[:,0]   
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Residual Total Forces
# ----------------------------------------------------------------------------------------------------------------------  
def residual_total_forces(segment):
    
    # Unpack results
    FT      = segment.state.conditions.frames.inertial.total_force_vector
    a       = segment.state.conditions.frames.inertial.acceleration_vector
    m       = segment.state.conditions.weights.total_mass    
    alt_in  = segment.state.unknowns.altitude[:,0] 
    alt_out = segment.state.conditions.freestream.altitude[:,0] 
    
    # Residual in X and Z, as well as a residual on the guess altitude
    if segment.flight_dynamics.force_x: 
        segment.state.residuals.force_x[:,0] = FT[:,0]/m[:,0] - a[:,0]
    if segment.flight_dynamics.force_y: 
        segment.state.residuals.force_y[:,0] = FT[:,1]/m[:,0] - a[:,1]       
    if segment.flight_dynamics.force_z: 
        segment.state.residuals.force_z[:,0] = FT[:,2]/m[:,0] - a[:,2]    
          
    segment.state.residuals.altitude[:,0] = (alt_in - alt_out)/alt_out[-1]

    return    

# ----------------------------------------------------------------------------------------------------------------------  
# Update Differentials
# ----------------------------------------------------------------------------------------------------------------------     
def update_differentials(segment):
    """ On each iteration creates the differentials and integration functions from knowns about the problem. 
      Sets the time at each point. Must return in dimensional time, with t[0] = 0.
      This is different from the common method as it also includes the scaling of operators.

        Assumptions:
        Works with a segment discretized in vertical position, altitude

        Inputs:
        state.numerics.dimensionless.control_points      [Unitless]
        state.numerics.dimensionless.differentiate       [Unitless]
        state.numerics.dimensionless.integrate           [Unitless]
        state.conditions.frames.inertial.position_vector [meter]
        state.conditions.frames.inertial.velocity_vector [meter/second]
        

        Outputs:
        state.conditions.frames.inertial.time            [second]

    """    

    # unpack
    numerics   = segment.state.numerics
    conditions = segment.state.conditions
    x          = numerics.dimensionless.control_points
    D          = numerics.dimensionless.differentiate
    I          = numerics.dimensionless.integrate 
    r          = segment.state.conditions.frames.inertial.position_vector
    v          = segment.state.conditions.frames.inertial.velocity_vector
    alt0       = segment.altitude_start
    altf       = segment.altitude_end    

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
        
    dz = altf - alt0
    vz = -v[:,2,None] # maintain column array

    # get overall time step
    dt = (dz/np.dot(I,vz))[-1]

    # rescale operators
    x = x * dt
    D = D / dt
    I = I * dt
    
    # Calculate the altitudes
    alt = np.dot(I,vz) + alt0
    
    # pack
    t_initial                                       = segment.state.conditions.frames.inertial.time[0,0]
    numerics.time.control_points                    = x
    numerics.time.differentiate                     = D
    numerics.time.integrate                         = I
    conditions.frames.inertial.time[1:,0]           = t_initial + x[1:,0]
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0]  
    conditions.freestream.altitude[:,0]             =  alt[:,0]  

    return